from bob.connections.air import AirBidirectionalConnectionPoint
from bob.connections.light import (
    LightVisibleInletConnectionPoint,
    LightVisibleOutletConnectionPoint,
)
from bob.core import BOB, S223, Equipment, PropertyReference

_namespace = BOB


class Window(Equipment):
    _class_iri = S223.Window
    indoor: AirBidirectionalConnectionPoint
    outdoor: AirBidirectionalConnectionPoint
    naturalLightInlet: LightVisibleInletConnectionPoint
    naturalLightOutlet: LightVisibleOutletConnectionPoint

    # Those will come from something else, but be accessible from here.
    openCloseStatus: PropertyReference
    openCloseCommand: PropertyReference
    shadeStatus: PropertyReference
    shadeCommand: PropertyReference
    breakDetection: PropertyReference



from typing import Dict

from ...connections.air import AirInletConnectionPoint, AirOutletConnectionPoint
from ...connections.electricity import (
    ElectricalInletConnectionPoint,
)
from ...connections.liquid import (
    HotWaterInletConnectionPoint,
    HotWaterOutletConnectionPoint,
    WaterOutletConnectionPoint,
)
from ...connections.naturalgas import NaturalGasInletConnectionPoint
from ...core import (
    BOB,
    S223,
    BoundaryConnectionPoint,
    Equipment,
)
from ...enum import Role
from ... import application
from ...template import SystemFromTemplate, template_update

_namespace = BOB

# if no configuration is passed, use this basic template of a generic hot water heater equipment
basic_hotwaterheater_template = {
    "params": {"label": "HotWaterHeater", "comment": "Hot Water Heater"},
    "equipment": {
        ("hw_heater", Equipment): {
            "config": {
                "cp": {
                    "hotWaterLeaving": HotWaterOutletConnectionPoint,
                    "hotWaterEntering": HotWaterInletConnectionPoint,
                    "electricalInlet": ElectricalInletConnectionPoint,
                }
            },
        },
    },
    "relations": [
        ("self.leavingFluid", "=", "self['hw_heater'].hotWaterLeaving"),
        ("self.enteringFluid", "=", "self['hw_heater'].hotWaterEntering"),
        ("self.electricalInlet", "=", "self['hw_heater'].electricalInlet"),
        (
            "self['hw_heater'].hotWaterLeaving",
            "**=",
            "self['hw_heater'].hotWaterEntering",
        ),
    ],
}


# 223 Standard Systems
class DomesticHotWaterHeater(SystemFromTemplate, application.HotWaterHeater):
    # _class_iri = S223.DomesticHotWaterHeater
    leavingFluid: BoundaryConnectionPoint
    enteringFluid: BoundaryConnectionPoint
    electricalInlet: BoundaryConnectionPoint

    def __init__(self, config: Dict = basic_hotwaterheater_template, **kwargs) -> None:
        _config = template_update({}, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        super().__init__(_config, **kwargs)
        self += Role.Heating
        self.leavingFluid.paired_to(self.enteringFluid)


# 223 Standard Equipment
class HotWaterBoiler(Equipment, application.Boiler):
    _class_iri = S223.Boiler
    hotWaterLeaving: HotWaterOutletConnectionPoint
    hotWaterEntering: HotWaterInletConnectionPoint

    def __init__(self, config: Dict = basic_hotwaterheater_template, **kwargs) -> None:
        _config = template_update({}, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        super().__init__(_config, **kwargs)
        self += Role.Heating
        self.hotWaterLeaving.paired_to(self.hotWaterEntering)


class ElectricalHotWaterBoiler(HotWaterBoiler, application.Boiler):
    _class_iri = S223.Boiler
    electricalInlet: ElectricalInletConnectionPoint


class NaturalGasHotWaterBoiler(HotWaterBoiler, application.Boiler):
    _class_iri = S223.Boiler
    electricalInlet: ElectricalInletConnectionPoint
    naturalGasInlet: NaturalGasInletConnectionPoint
    combustionAirInlet: AirInletConnectionPoint
    combustionAirOutlet: AirOutletConnectionPoint
    condensedWaterOutlet: WaterOutletConnectionPoint



from typing import Dict

from ...equipment.electricity import _MotorStarter
from ...properties.states import OnOffCommand

from ...connections import electricity as elec_cnx
from ...core import (
    BOB,
    P223,
    logging,
)
from ...template import template_update, configure_relations

# logging
_log = logging.getLogger(__name__)

# namespace
_namespace = BOB

electric_starter_template = {
    "cp": {
        "electricalInlet": elec_cnx.InletConnectionPoint,
        "electricalOutlet": elec_cnx.OutletConnectionPoint,
    },
    "properties": {
        # ("actuatesProperty", PercentCommand): {},
        ("onOffCommand", OnOffCommand): {},
    },
}


class MotorStarter(_MotorStarter):
    """
    Motor starter
    This Equipment provides command and status for an electrical
    Equipment like a fan or a pump

    """

    _class_iri = P223.MotorStarter

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update(electric_starter_template, config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _log.debug(f"MotorStarter.__init__ {_config} {kwargs}")
        _relations = _config.pop("relations", [])
        super().__init__(_config, **kwargs)
        configure_relations(self, _config.pop("relations", []))



import copy
import re
import typing as t
from pathlib import Path

import yaml

from .core import Connection, Equipment, System
from .introspection import get_class_from_name


def template_update(base: t.Dict = {}, config: t.Dict = None, bases: t.List = None):
    """
    This utility allows to preserve module templates from
    undesired modification during creation of Equipment.

    Usage :
    _config = template_update(template, user_provided_config_dict)

    """

    def merge_dict(existing, new):
        for k in new:
            if k in existing:
                if isinstance(existing[k], dict) and isinstance(new[k], dict):
                    merge_dict(existing[k], new[k])
                else:
                    existing[k] = new[k]
            else:
                existing[k] = new[k]

    if bases:
        d1, d2 = bases
        _d1 = copy.deepcopy(d1)
        _d2 = copy.deepcopy(d2)
        merge_dict(_d1, _d2)
        if config:
            merge_dict(_d1, config)
        return _d1

    else:
        _d = copy.deepcopy(base)
        if config:
            merge_dict(_d, config)
        return _d


def get_instance(container: t.Union[Equipment, System], blob: str):
    # print('Looking for : ', container, blob)
    if "[" in blob:
        matches = re.findall(r'\[["\'](.*?)["\']\]', blob)  # sub-equipment
        property_match = re.search(
            r"\.(?P<property>\w+)$", blob
        )  # property => .something
        thing = container[matches.pop(0)]

        for each in matches:
            thing = thing[each]
        # print('thing : ', thing, property_match)
        if property_match:
            property_name = property_match.group("property")
            # try:
            #
            #    thing_property = getattr(thing, property_name)
            #    if thing_property is None:
            #        thing_property = thing # in case thing_property is None, we give the part before .something
            # except AttributeError:
            #    thing_property = None
            # print(thing_property, property_name)
            return (thing, property_name)  # in case thing_property is None
        # print(thing, None)
        return (thing, None)
    else:
        _key = blob.split(".")[1]
        thing = getattr(container, _key)
        # print(thing, _key)
        return (thing, _key)  # in case thing is None


def configure_relations(
    container: t.Union[Equipment, System], relations: t.List[t.Tuple[str, str, str]]
):
    for relation in relations:
        _source, operator, _target = relation
        source_element, source_key = get_instance(container, _source)
        target_element, target_key = get_instance(container, _target)

        if source_key is None:
            source = source_element
        else:
            source = getattr(source_element, source_key, None)
        if source is None and isinstance(source_element, Connection):
            source = source_element

        # if source_element is None:
        #    source_element = container

        # if source_key is None:
        #    source = source_element
        # else:
        #    source = getattr(source_element, source_key, None)

        # if source is None and isinstance(source_element, Connection):
        #    source = source_element
        # elif source is None:
        #    raise AttributeError(f"Source {source_key} not found in {source_element} | container {container} | relation {relation}")

        if target_key is None:
            target = target_element
        else:
            target = getattr(target_element, target_key, None)

        if target is None and isinstance(target_element, Connection):
            target = target_element
        elif target is None:
            raise AttributeError(f"Target {target_key} not found in {target_element}")

        if operator == "=":
            if source is None:
                # print(equipment, source_key, target)
                try:
                    setattr(source_element, source_key, target)
                except AttributeError:
                    setattr(container, source_key, target)
                except TypeError as error:
                    print(error)
                    print("Container :", container)
                    print("Source :", source, source_key)
                    print("Target :", target, target_key)

            else:
                source = target
        elif operator == ">>":
            source >> target
        elif operator == "<<":
            source << target
        elif operator == "%":
            source % target
        elif operator == "mapsTo":
            source.mapsTo = target
        elif operator == "@":
            source @ target
        # no @ here as we are creating relation "inside" the equipment or system


class SystemFromTemplate(System):
    def __init__(self, config: t.Dict = None, **kwargs):
        _config = template_update(config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _relations = _config.pop("relations", [])
        super().__init__(_config, **kwargs)
        configure_relations(self, _relations)


def config_from_yaml(yaml_file: t.Union[str, Path] = None):
    if yaml_file is None:
        raise FileNotFoundError("No YAML file provided")
    else:
        yaml_file = Path(yaml_file)
    with open(yaml_file, "r") as file:
        yaml_content = yaml.safe_load(file)
    _text_values = ["label", "comment"]
    _dict = {}
    name = yaml_content["name"]
    params = yaml_content["params"]
    label = params.get("label", name)
    comment = params.get("comment", "")
    sensors = yaml_content.get("sensors", None)
    equipment = yaml_content.get("equipment", None)

    _dict["params"] = {"label": label, "comment": comment}

    def define_entities(entities: dict = None, entities_category: str = None):
        if entities is None:
            return
        _dict[entities_category] = {}
        for entity_name, entity_params in entities.items():
            # entity_label = entity_params['label'] if 'label' in entity_params else entity_name
            entity_label = entity_params.pop("label", entity_name)
            # print(entity_label, entity_params, f"Looking for {entity_params['class']}")
            entity_class = get_class_from_name(entity_params.pop("class"))
            # print('Found class', entity_class)
            # entity_comment = entity_params.pop('comment', '')
            _dict[entities_category][(entity_label, entity_class)] = {}
            for _name, _class_or_value in entity_params.items():
                _value = (
                    _class_or_value
                    if _name in _text_values
                    else get_class_from_name(_class_or_value)
                )
                _dict[entities_category][(entity_label, entity_class)][_name] = _value

    # print('Defining entities')
    define_entities(equipment, "equipment")
    define_entities(sensors, "sensors")
    _dict["relations"] = []
    _relations = yaml_content.get("relations", [])
    for _relation in _relations:
        _relation = _relation.replace("(", "").replace(")", "").strip()
        _a, _b, _c = _relation.split(",")
        _dict["relations"].append((_a.strip(), _b.strip(), _c.strip()))
    return _dict



from ...connections.air import AirInletConnectionPoint, AirOutletConnectionPoint
from ...core import BOB, S223, Equipment
from ...sensor import Sensor

_namespace = BOB


class AirFlowMonitor(Equipment):
    _class_iri = S223.FlowSensor
    airInlet: AirInletConnectionPoint
    airOutlet: AirOutletConnectionPoint
    flowSensor: Sensor


# TODO : Create the template and make that the same than the others.



from rdflib import Literal, URIRef

from ..core import (
    BOB,
    S223,
    EnumerationKind,
    QuantifiableProperty,
)

_namespace = BOB


class Setpoint(QuantifiableProperty):
    _class_iri: URIRef = S223.Setpoint
    hasAspect: EnumerationKind
    hasDeadband: Literal
    hasValue: Literal
    hasQuantityKind: URIRef
    hasUnit: URIRef

    def __init__(self, **kwargs):
        _properties = {}
        for k, v in self.__annotations__.items():
            if k in kwargs:
                _properties[k] = kwargs.pop(k)
        super().__init__(**kwargs)
        for k, v in _properties.items():
            if v is not None:
                setattr(self, k, self.__annotations__[k](v))



import logging
from typing import Dict

from ..core import (
    G36,
    DomainSpace,
    Zone,
)
from ..sensor.flow import FlowSetpoint
from ..sensor.temperature import TemperatureSetpoint
from ..template import template_update
from . import (
    Function,
    FunctionInput,
    G36AnalogInput,
    G36AnalogOutput,
    G36DigitalInput,
)

# logging
_log = logging.getLogger(__name__)

# namespace
_namespace = G36


class G36Sequence(Function):
    """
    This function is a subclass of a Function Block kept
    in the namespace of G36.

    In Guideline 36, models present the notion of AI, AO, BI, BO
    and those concept can be modeled using a Function block.
    Function block is then an abstraction of the sequence of
    operation suggested by G36.

    Comment of this block is the description of the sequence.
    """

    _class_iri = G36.Function


class G36ZoneTemperatureControl(G36Sequence):
    _class_iri = G36.Function


class G36AirFlowControl(G36Sequence):
    _class_iri = G36.Function


class G36OccupancyControl(G36Sequence):
    _class_iri = G36.Function


class G36VentilationAndCO2Control(G36Sequence):
    _class_iri = G36.Function


# g36_4-1_VAV_TerminalUnit_CoolingOnly
class G36VAVCoolingOnly(G36Sequence):
    _class_iri = G36.VAVCoolingOnly
    # From table in section 4.1
    # Defined as Function Input and Output as we will connect
    # to existing properties in the model
    boxDamperPosition: G36AnalogOutput
    supplyAirFlow: G36AnalogInput
    zoneTemperature: G36AnalogInput
    localOverride: G36DigitalInput
    zoneOccupancySensor: G36DigitalInput
    zonewindowSwitch: G36DigitalInput
    zoneSetpointAdj: G36AnalogInput
    zoneCO2: G36AnalogInput
    effectiveOccupancy: FunctionInput
    ahuSupplyAirTemp: FunctionInput

    ### PROPERTIES
    # Airflow setpoints
    zoneMaximumCoolingAirflowSetpoint: FlowSetpoint
    zoneMaximumHeatingAirflowSetpoint: FlowSetpoint

    # Temperature setpoint
    occupiedHtgSetpoint: TemperatureSetpoint
    occupiedClgSetpoint: TemperatureSetpoint
    unoccupiedHtgSetpoint: TemperatureSetpoint
    unoccupiedClgSetpoint: TemperatureSetpoint

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update({}, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _log.debug(f"G36VAVCoolingOnly.__init__ {_config} {kwargs}")

        super().__init__(**kwargs)


class G36Figure_A_3(G36VAVCoolingOnly):
    def __init__(self, comment=None, **kwargs):
        super().__init__(comment=comment, **kwargs)


class G36Figure_A_2(G36VAVCoolingOnly):
    def __init__(self, comment=None, **kwargs):
        super().__init__(comment=comment, **kwargs)


class G36VAVCoolingOnly0(G36Sequence):
    def __init__(self, comment=None, **kwargs):
        super().__init__(comment=comment, **kwargs)

    _class_iri = G36.Function


class G36ZoneGroup(Zone):
    _class_iri = G36.ZoneGroup


class G36ZoneFromZone(Zone):
    _class_iri = G36.Zone


class G36ZoneFromDomainSpace(DomainSpace):
    _class_iri = G36.Zone


VAV_CoolingOnly_template = {
    "cp": {
        "boxDamperPosition": G36AnalogOutput,
        "supplyAirFlow": G36AnalogInput,
        "zoneTemperature": G36AnalogInput,
        "localOverride": G36DigitalInput,
        "zoneOccupancySensor": G36DigitalInput,
        "zonewindowSwitch": G36DigitalInput,
        "zoneSetpointAdj": G36AnalogInput,
        "zoneCO2": G36AnalogInput,
        "effectiveOccupancy": FunctionInput,
        "ahuSupplyAirTemp": FunctionInput,
    },
    "functions": {
        ("zoneTemperatureControl", G36ZoneTemperatureControl): {},
        ("airFlowControl", G36AirFlowControl): {},
        ("occupancyControl", G36OccupancyControl): {},
        ("ventilationAndCO2Control", G36VentilationAndCO2Control): {},
    },
}



from typing import Any, Tuple


from bob.producer.causality import Differential
from bob.properties.force import DifferentialStaticPressure, Pressure

from ..core import (
    BOB,
    INCLUDE_INVERSE,
    S223,
    Node,
    PropertyReference,
)
from ..enum import Air, Water
from ..properties import DifferentialStaticPressure
from .sensor import Sensor, split_kwargs

_namespace = BOB  #


class PressureSensor(Sensor):
    _class_iri = S223.PressureSensor
    observes: PropertyReference  # Temperature
    # hasObservationLocation: LocationReference

    def __init__(self, **kwargs: Any) -> None:
        _sensor_kwargs, _property_kwargs = split_kwargs(kwargs)

        if "hasUnit" not in _property_kwargs:
            raise ValueError("You must provide hasUnit when defining a pressure sensor")
        if "ofMedium" not in _property_kwargs:
            raise ValueError(
                "You must provide ofMedium when defining a pressure sensor"
            )

        super().__init__(**_sensor_kwargs)

        self.observes = Pressure(
            # isObservedBy=self,
            label=f"{self.label}.GaugePressure",
            **_property_kwargs,
        )


# class DifferentialStaticPressureSetpoint(Setpoint):
#    _class_iri = S223.Sensor
#    hasQuantityKind: URIRef = QUANTITYKIND.ForcePerArea
#    hasUnit: URIRef


class DifferentialStaticPressureSensor(Sensor):
    _class_iri = S223.PressureSensor
    observes: PropertyReference
    observation_pressure: Pressure
    reference_pressure: Pressure
    differential_static_pressure: DifferentialStaticPressure
    highPort: PressureSensor
    lowPort: PressureSensor

    def __init__(self, **kwargs: Any) -> None:
        _sensor_kwargs, _property_kwargs = split_kwargs(kwargs)

        super().__init__(**_sensor_kwargs)

    def add_hasObservationLocation(self, node: Tuple[Node, Node]) -> None:
        # For now, make that a secret, or we end up with s223.hasObservationLocation
        # self._hasObservationLocation = node

        # link the two together
        observation_location, reference_location = node
        self._data_graph.add(
            (self._node_iri, S223.hasReferenceLocation, reference_location._node_iri)
        )
        if INCLUDE_INVERSE:
            reference_location.isReferenceLocation = self

        self._data_graph.add(
            (
                self._node_iri,
                S223.hasObservationLocation,
                observation_location._node_iri,
            )
        )
        if INCLUDE_INVERSE:
            observation_location.isReferenceLocation = self

        self["highPort"] % observation_location
        self["lowPort"] % reference_location


class AirDifferentialStaticPressureSensor(DifferentialStaticPressureSensor):
    _class_iri = S223.PressureSensor
    # observes: PropertyReference

    def __init__(self, **kwargs):
        _sensor_kwargs, _property_kwargs = split_kwargs(kwargs)
        super().__init__(**_sensor_kwargs)
        self.differential_static_pressure = DifferentialStaticPressure(
            ofMedium=Air,
            label=f"{self.label}.DifferentialStaticPressure",
            **_property_kwargs,
        )
        self.observation_pressure = Pressure(
            ofMedium=Air,
            label=f"{self.label}.ObservationPressure",
            **_property_kwargs,
        )
        self.reference_pressure = Pressure(
            ofMedium=Air,
            label=f"{self.label}.ReferencePressure",
            **_property_kwargs,
        )
        self > Differential(
            label="diff_causality", comment="Will output High minus Low"
        )
        self > PressureSensor(label="highPort", ofMedium=Water, **_property_kwargs)
        self > PressureSensor(label="lowPort", ofMedium=Water, **_property_kwargs)
        self > Differential(label="output", comment="Will output High minus Low")
        # self["highPort"].observes >> self.observation_pressure
        # self["lowPort"].observes >> self.reference_pressure
        self.observation_pressure >> self["diff_causality"].high_input
        self.reference_pressure >> self["diff_causality"].low_input
        self["diff_causality"].differential_output >> self.differential_static_pressure
        self.observes = self.differential_static_pressure


class WaterDifferentialStaticPressureSensor(DifferentialStaticPressureSensor):
    _class_iri = S223.DifferentialSensor

    def __init__(self, **kwargs):
        _sensor_kwargs, _property_kwargs = split_kwargs(kwargs)
        super().__init__(**_sensor_kwargs)
        self.differential_static_pressure = DifferentialStaticPressure(
            ofMedium=Water,
            label=f"{self.label}.DifferentialStaticPressure",
            **_property_kwargs,
        )
        self.observation_pressure = Pressure(
            ofMedium=Water,
            label=f"{self.label}.ObservationPressure",
            **_property_kwargs,
        )
        self.reference_pressure = Pressure(
            ofMedium=Water,
            label=f"{self.label}.ReferencePressure",
            **_property_kwargs,
        )
        self > Differential(
            label="diff_causality", comment="Will output High minus Low"
        )
        self > PressureSensor(label="highPort", ofMedium=Water, **_property_kwargs)
        self > PressureSensor(label="lowPort", ofMedium=Water, **_property_kwargs)
        self > Differential(label="output", comment="Will output High minus Low")
        # self["highPort"].observes >> self.observation_pressure
        # self["lowPort"].observes >> self.reference_pressure
        self.observation_pressure >> self["diff_causality"].high_input
        self.reference_pressure >> self["diff_causality"].low_input
        self["diff_causality"].differential_output >> self.differential_static_pressure
        self.observes = self.differential_static_pressure



import logging
from typing import Dict


from ... import application
from ...core import BOB, BoundaryConnectionPoint, System
from ...template import SystemFromTemplate, template_update

# logging
_log = logging.getLogger(__name__)

# namespace
_namespace = BOB

ahu_template = {
    "params": {},
    "sensors": {},
    "equipment": {},
}

fan_coil_template = {
    "params": {},
    "sensors": {},
    "equipment": {},
}


class AirHandlingUnit(SystemFromTemplate, application.AirHandlingUnit):
    """
        This is treated as a system.
        We will add multiple devices inside using a template

        example from pritoni (2024-01-06)
        ahu_template = {
        "params": {"label": "AHU", "comment": "AHU delivering air to 2 VAV boxes"},
        "sensors": {
            ("OA-T", AirTemperatureSensor): {
                "hasUnit": UNIT.DEG_C,
                "comment": "Oudoor air temperature (S3)",
            },
            ("TPD1", AirDifferentialStaticPressureSensor): {
                "hasUnit": UNIT.PA,
                "comment": "Filter Differential Pressure Sensor (S5)",
            },
            ("HC-T", AirTemperatureSensor): {
                "hasUnit": UNIT.DEG_C,
                "comment": "Air temperature after heating coil (S6)",
            },
            ("MA-T", AirTemperatureSensor): {
                "hasUnit": UNIT.DEG_F,
                "comment": "Return Air temperature (S4)",
            },
            ("DA-T", AirTemperatureSensor): {
                "hasUnit": UNIT.DEG_F,
                "comment": "Discharge Air temperature after cooling coil (S7)",
            },
            ("RA-T", AirTemperatureSensor): {
                "hasUnit": UNIT.DEG_F,
                "comment": "Return Air temperature (S2)",
            },
            ("TPD2", AirDifferentialStaticPressureSensor): {
                "hasUnit": UNIT.PA,
                "comment": "Supply Duct Static Pressure (S8)",
            },
            ("TPD3", AirDifferentialStaticPressureSensor): {
                "hasUnit": UNIT.PA,
                "comment": "Return Duct Static Pressure (S1)",
            },
        },
        "equipment": {
            ("RF", Fan): {
                "comment": "Return Air Fan",
                "electricalInlet": Electricity_600VLL_3Ph_60HzInletConnectionPoint,
                "hasRole": Role.Return,
            },
            ("RF_VFD", VFD): {
                "comment": "Return Air Fan VFD",
            },
            ("SF", Fan): {
                "comment": "Supply Air Fan",
                "electricalInlet": Electricity_600VLL_3Ph_60HzInletConnectionPoint,
                "hasRole": Role.Supply,
            },
            ("SF_Starter", MotorStarter): {
                "comment": "Supply Air Fan Starter",
            },
            ("CLGCOIL", ChilledWaterCoil): {"comment": "Cooling Coil"},
            ("HTGCOIL", HotWaterCoil): {"comment": "Heating coil"},
            ("FILTER", Filter): {"comment": "Filter"},
            ("OADPR", ElectricalActuatedProportionalDamper): {
                "comment": "Outdoor air damper (A3)"
            },
            ("MADPR", ElectricalActuatedProportionalDamper): {
                "comment": "Mixed Air Damper (A2)"
            },
            ("EADPR", ElectricalActuatedProportionalDamper): {
                "comment": "Exhaust Air Damper (A1)"
            },
        },
    }

    """

    _class_iri = BOB.AirHandlingUnit
    outsideAirInlet: BoundaryConnectionPoint
    returnAirInlet: BoundaryConnectionPoint
    supplyAirOutlet: BoundaryConnectionPoint
    exhaustAirOutlet: BoundaryConnectionPoint
    electricalInlet: BoundaryConnectionPoint
    # The system connection points are now a Bob thing only, there is no such thing in 223P anymore

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update(ahu_template, config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _log.debug(f"AirHandlingUnit.__init__ {_config} {kwargs}")
        super().__init__(_config, **kwargs)


class FanCoil(System, application.FanCoilUnit):
    _class_iri = BOB.FanCoilUnit
    returnAirInlet: BoundaryConnectionPoint
    supplyAirOutlet: BoundaryConnectionPoint
    exhaustAirOutlet: BoundaryConnectionPoint
    electricalInlet: BoundaryConnectionPoint

    def __init__(self, config: Dict = {}, **kwargs) -> None:
        _config = template_update(fan_coil_template, config)
        kwargs = {**_config.get("params", {}), **kwargs}
        _log.debug(f"FanCoil.__init__ {_config} {kwargs}")
        super().__init__(config, **kwargs)
        # self += SystemType.FanCoilUnit



import importlib
import inspect
import pkgutil
import typing as t

import bob

class_cache = {}
module_cache = {}
enum_cache = {}


def look_in_cache(name: str = None, cache: dict = None):
    return cache[name] if name in cache else None


def get_modules_from(package_name: str) -> t.List[str]:
    package = importlib.import_module(package_name)
    modules = []
    for importer, modname, ispkg in pkgutil.walk_packages(
        package.__path__, package.__name__ + "."
    ):
        modules.append(modname)
    return modules


def load_modules(module: t.Type = bob, force: bool = False) -> None:
    global class_cache
    global module_cache
    # print(f"Loading modules {module}")
    if force or module.__name__ not in module_cache:
        module_cache[module.__name__] = get_modules_from(module.__name__)


def get_class_from_name(classname: str = None, module: t.Type = bob) -> t.Type:
    global class_cache
    global module_cache
    super = None
    if not classname:
        raise ValueError("Classname is required")

    _existing = look_in_cache(classname, class_cache)
    if _existing:
        return _existing

    if "." in classname:
        super, classname = classname.split(".")
    if "|" in classname:
        _module, classname = classname.split("|")
        module = importlib.import_module(_module)
    load_modules(module)
    for each in module_cache[module.__name__]:
        try:
            submodule = importlib.import_module(f"{each}")
        except ImportError:
            continue
        for name, obj in inspect.getmembers(submodule):
            if (
                isinstance(obj, bob.core.EnumerationKind)
                and "bob.enum" in submodule.__name__
            ):
                _class = getattr(importlib.import_module(f"{submodule.__name__}"), name)
                enum_cache[name] = _class

            if inspect.isclass(obj):
                if name == classname:
                    if super is not None:
                        _key = f"{super}.{classname}"
                    else:
                        _key = classname
                    class_cache[_key] = obj
                    return obj
                else:
                    if name not in class_cache:
                        class_cache[name] = obj
        if super is not None:
            if super in enum_cache:
                _super_class = enum_cache[super]
                return getattr(_super_class, classname)
            else:
                if classname in enum_cache:
                    return enum_cache[classname]
    raise TypeError(
        f"Class {super} {classname} not found in {module}, cache is {class_cache}"
    )



from typing import Dict

from rdflib import URIRef

from ...connections.air import (
    CompressedAirInletConnectionPoint,
    CompressedAirOutletConnectionPoint,
)
from ...connections.liquid import WaterInletConnectionPoint, WaterOutletConnectionPoint
from ...connections.mechanical import MechanicalInletConnectionPoint
from ...connections.naturalgas import (
    NaturalGasInletConnectionPoint,
    NaturalGasOutletConnectionPoint,
)
from ...connections.refrigerant import (
    RefrigerantBidirectionalConnectionPoint,
    RefrigerantInletConnectionPoint,
    RefrigerantOutletConnectionPoint,
)
from ...core import BOB, S223, Equipment, PropertyReference, logging
from ...enum import (  # , R134a, R404a, R407c, R448a, R449a, R452a, R454b, R507a
    Fluid,
    Refrigerant,
)
from ...properties import Gallons
from ...template import configure_relations, template_update

# logging
_log = logging.getLogger(__name__)

# namespace
_namespace = BOB


class Valve(Equipment):
    """
    Base class for a valve. Must be subclassed to provide inlet and outlet
    depending on configuration
    """

    _class_iri: URIRef = S223.Valve
    linkageInlet: MechanicalInletConnectionPoint
    position: PropertyReference
    command: PropertyReference
    position_feedback: PropertyReference
    flowCoefficient: Gallons


class TwoWayValve(Valve):
    """
    Two-way valve have 1 inlet and 1 outlet
    """

    _class_iri: URIRef = S223.TwoWayValve
    fluidInlet: WaterInletConnectionPoint
    fluidOutlet: WaterOutletConnectionPoint
    is_open: PropertyReference
    is_closed: PropertyReference

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update({}, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _relations = _config.pop("relations", [])
        super().__init__(_config, **kwargs)
        configure_relations(self, _relations)
        self.fluidOutlet.paired_to(self.fluidInlet)

    def set_fluid_type(self, fluid: Fluid):
        self.set_medium(["fluidInlet", "fluidOutlet"], fluid)


class ThreeWayValveDiverting(Valve):
    """
    A diverting valve has 1 inlet and 2 outlets
    """

    _class_iri = S223.ThreeWayValve
    fluidInletAB: WaterInletConnectionPoint
    fluidOutletA: WaterOutletConnectionPoint
    fluidOutletB: WaterOutletConnectionPoint

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update({}, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _relations = _config.pop("relations", [])
        super().__init__(_config, **kwargs)
        configure_relations(self, _relations)
        self.fluidOutletA.paired_to(self.fluidInletAB)
        self.fluidOutletB.paired_to(self.fluidInletAB)

    def set_fluid_type(self, fluid: Fluid):
        self.set_medium(["fluidInletAB", "fluidOutletA", "fluidOutletB"], fluid)


class ThreeWayValveMixing(Valve):
    """
    A mixing valve has 2 inlet and 1 outlet
    """

    _class_iri: URIRef = S223.ThreeWayValve
    fluidInletA: WaterInletConnectionPoint
    fluidInletB: WaterOutletConnectionPoint
    fluidOutlet: WaterOutletConnectionPoint

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update({}, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _relations = _config.pop("relations", [])
        super().__init__(_config, **kwargs)
        configure_relations(self, _relations)
        self.fluidOutlet.paired_to(self.fluidInletA)
        self.fluidOutlet.paired_to(self.fluidInletB)

    def set_fluid_type(self, fluid: Fluid):
        self.set_medium(["fluidInletA", "fluidInletB", "fluidOutlet"], fluid)


class NaturalGasValve(Valve):
    _class_iri = S223.Valve
    naturalGasInlet: NaturalGasInletConnectionPoint
    naturalGasOutlet: NaturalGasOutletConnectionPoint

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update({}, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _relations = _config.pop("relations", [])
        super().__init__(_config, **kwargs)
        configure_relations(self, _relations)
        self.naturalGasOutlet.paired_to(self.naturalGasInlet)


class PneumaticValve(Valve):
    _class_iri = S223.Valve
    compressedAirInlet: CompressedAirInletConnectionPoint
    compressedAirOutlet: CompressedAirOutletConnectionPoint

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update({}, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _relations = _config.pop("relations", [])
        super().__init__(_config, **kwargs)
        configure_relations(self, _relations)
        self.compressedAirOutlet.paired_to(self.compressedAirInlet)


class ExpansionValve(Valve):
    _class_iri = S223.Valve
    portA: RefrigerantBidirectionalConnectionPoint
    portB: RefrigerantBidirectionalConnectionPoint

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update({}, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _relations = _config.pop("relations", [])
        super().__init__(_config, **kwargs)
        configure_relations(self, _relations)
        self.portB.paired_to(self.portA)

    def set_gas_type(self, gas: Refrigerant):
        self.set_medium(["portA", "portB"], gas)


class ReversingValve(Valve):
    _class_iri = S223.Valve
    refrigerantHighPressureInlet: RefrigerantInletConnectionPoint
    refrigerantLowPressureOutlet: RefrigerantOutletConnectionPoint
    refrigerantIndoorCoilPort: RefrigerantBidirectionalConnectionPoint
    refrigerantOutdoorCoilPort: RefrigerantBidirectionalConnectionPoint
    position: PropertyReference

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update({}, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _relations = _config.pop("relations", [])
        super().__init__(_config, **kwargs)
        configure_relations(self, _relations)
        # self.refrigerantLowPressureOutlet.paired_to(self.refrigerantHighPressureInlet)

    def set_gas_type(self, gas: Refrigerant):
        self.set_medium(
            [
                "refrigerantHighPressureInlet",
                "refrigerantLowPressureOutlet",
                "refrigerantIndoorCoilPort",
                "refrigerantOutdoorCoilPort",
            ],
            gas,
        )



# flake8: noqa

#
#   Project Metadata
#

__version__ = "0.99"

# core classes for S223 models
# basic pieces for all Bob models
from .core import DomainSpace  # portion of a physical space for a specific domain
from .core import Equipment  # a piece of equipment
from .core import Node  # a node in a graph
from .core import PhysicalSpace  # heirarchy of things like building, floor, room
from .core import System  # a collection of equipment
from .core import Zone  # collections of domain spaces
from .core import bind_model_namespace  # associate a namespace and prefix
from .core import data_graph  # the data model created
from .core import dump  # output a graph, defaults to data_graph
from .core import schema_graph  # the schema for the custom components in the model

# from . import enum



from rdflib import URIRef

from ..core import (
    BOB,
    QUANTITYKIND,
    Medium,
    QuantifiableObservableProperty,
)

_namespace = BOB


class Temperature(QuantifiableObservableProperty):
    hasQuantityKind = QUANTITYKIND.Temperature
    hasUnit: URIRef
    ofMedium: Medium  # set from the sensor
    # isObservedBy: Sensor



from pyclbr import Function
from typing import Any, Dict

from rdflib import URIRef

from ...multimethods import multimethod
from ...producer import Function

from ...connections.electricity import Electricity_24VLN_1Ph_60HzInletConnectionPoint
from ...connections.network import RS485BidirectionalConnectionPoint
from ...core import (
    INCLUDE_INVERSE,
    P223,
    S223,
    Equipment,
    data_graph,
    logging,
)

from ...externalreference import NetworkProfile
from ...template import template_update
from . import AnalogInput, AnalogOutput

# logging
_log = logging.getLogger(__name__)

# namespace
_namespace = P223

# empty, because other equipments can be defined as controller
# and I don't want them to inherit from example properties
controller_template = {
    "cp": {},
    "properties": {},
}

controller_template_example = {
    "cp": {
        "electricalInlet": Electricity_24VLN_1Ph_60HzInletConnectionPoint,
        "bacnet_mstp": RS485BidirectionalConnectionPoint,
        "zone_temperature_sensor": AnalogInput,
        "airflow_sensor": AnalogInput,
        "damper_output": AnalogOutput,
    },
    "properties": {},
}


class Controller(Equipment):
    """
    A controller executes function blocks and connect to other Equipment
    through different connection points (AI, AO, BI, BO)
    """

    _class_iri: URIRef = S223.Controller
    _attr_uriref = {"hasNetworkProfile": P223.hasNetworkProfile}
    # electricalInlet: Electricity_24VLN_1Ph_60HzInletConnectionPoint
    # executes: Function
    hasNetworkProfile: NetworkProfile

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update(controller_template, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _log.debug(f"Controller.__init__ {_config} {kwargs}")

        super().__init__(_config, **kwargs)

    def executes(self, function_block: Function):
        _log.debug(f"Controller {self._node_iri} executes  {function_block._node_iri}")
        data_graph.add((self._node_iri, S223.executes, function_block._node_iri))

    def __rshift__(self, other: Any) -> Any:
        """Build a connection from this thing to another thing."""
        connect_mm(self, other)
        return other

    def __lshift__(self, other: Any) -> Any:
        """Build a connection to this thing from another thing."""
        connect_mm(other, self)
        return self


@multimethod
def connect_mm(controller: Controller, network_Equipment: NetworkProfile) -> None:
    """Controller >> NetworkProfile"""
    _log.info(f"connect from {controller} to {network_Equipment}")

    data_graph.add(
        (controller._node_iri, P223.hasNetworkProfile, network_Equipment._node_iri)
    )
    if INCLUDE_INVERSE:
        data_graph.add(
            (network_Equipment._node_iri, P223.isNetworkProfileOf, controller._node_iri)
        )



from bob.core import P223, S223, Equipment, Property
from bob.equipment.control.controller import Controller

_namespace = S223


class _MotorStarter(Equipment):
    """
    This is required here so actuatesProperty gets its namespace from S223
    """

    _class_iri = S223.Equipment
    actuatesProperty: Property


class _VFD(Controller):
    """
    This is required here so actuatesProperty gets its namespace from S223
    """

    _class_iri = S223.VariableFrequencyDrive
    actuatesProperty: Property



from typing import Dict

from rdflib import URIRef

from ...equipment.electricity import _VFD

from ...connections.electricity import (
    ElectricalInletConnectionPoint,
    ElectricalOutletConnectionPoint,
)
from ...core import (
    S223,
    SCRATCH,
    logging,
)
from ...template import template_update, configure_relations

# logging
_log = logging.getLogger(__name__)

# namespace
_namespace = SCRATCH


vfd_template = {
    "cp": {
        "electricalInlet": ElectricalInletConnectionPoint,
        "electricalOutlet": ElectricalOutletConnectionPoint,
    },
    "properties": {},
    "parts": {},
}


class VFD(_VFD):
    _class_iri: URIRef = S223.VariableFrequencyDrive

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update(vfd_template, config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _log.debug(f"VFD.__init__ {_config} {kwargs}")
        _relations = _config.pop("relations", [])
        super().__init__(_config, **kwargs)
        configure_relations(self, _relations)



from bob.connections.light import LightVisibleInletConnectionPoint
from bob.properties.states import OccupancyStatus

from ..core import BOB, Domain, DomainSpace, PropertyReference, Zone

_namespace = BOB


class LightingSpace(DomainSpace):
    _class_iri = BOB.LightingSpace
    hasDomain = Domain.Lighting
    lightInlet: LightVisibleInletConnectionPoint
    naturalLightInlet: LightVisibleInletConnectionPoint
    occupancy: OccupancyStatus


class LightingZone(Zone):
    _class_iri = BOB.LightingZone
    hasDomain = Domain.Lighting
    # lightInlet: LightVisibleInletZoneConnectionPoint
    occupancy: PropertyReference  # promoted from a space



from typing import Dict

from rdflib import URIRef

from bob.equipment.electricity.starter import MotorStarter
from bob.equipment.electricity.vfd import VFD
from bob.properties.ratio import PercentCommand

from ...connections.electricity import (
    Electricity_600VLL_3Ph_60HzInletConnectionPoint,
    Electricity_600VLL_3Ph_60HzOutletConnectionPoint,
)
from ...connections.liquid import WaterInletConnectionPoint, WaterOutletConnectionPoint
from ...core import BOB, S223, UNIT, Equipment, PropertyReference, logging
from ...properties import (
    HP,
    RPM,
    Amps,
    ElectricPowerkW,
    PowerFactor,
    Pressure,
)
from ...template import template_update

# logging
_log = logging.getLogger(__name__)

# namespace
_namespace = BOB

pump_template = {
    "cp": {"electricalInlet": Electricity_600VLL_3Ph_60HzInletConnectionPoint},
    "properties": {
        ("head_pressure", Pressure): {"hasUnit": UNIT.PSI},
        ("amps", Amps): {},
        ("rpm", RPM): {},
        ("hp", HP): {},
        ("kW", ElectricPowerkW): {},
        ("powerFactor", PowerFactor): {},
    },
}


class Pump(Equipment):
    _class_iri = S223.Pump
    fluidInlet: WaterInletConnectionPoint
    fluidOutlet: WaterOutletConnectionPoint
    onOffStatus: PropertyReference
    onOffCommand: PropertyReference

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update(pump_template, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _log.debug(f"Fan.__init__ {_config} {kwargs}")

        super().__init__(_config, **kwargs)
        self.fluidOutlet.paired_to(self.fluidInlet)


starter_addon_template = {
    "equipment": {
        ("starter", MotorStarter): {
            "config": {
                "cp": {
                    "electricalInlet": Electricity_600VLL_3Ph_60HzInletConnectionPoint,
                    "electricalOutlet": Electricity_600VLL_3Ph_60HzOutletConnectionPoint,
                }
            },
        }
    },
    "properties": {("speedRatio", PercentCommand): {}},
}


class PumpWithStarter(Pump):
    """
    This pump is composed of a pump, an electrical motor and a starter
    """

    _class_iri: URIRef = S223.Pump

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update(
            bases=[pump_template, starter_addon_template],
            config=config,
        )
        kwargs = {**_config.pop("params", {}), **kwargs}
        _log.debug(f"PumpWithStarter.__init__ {_config} {kwargs}")

        super().__init__(_config, **kwargs)
        self.onOffCommand = self["starter"]["onOffCommand"]
        self.onOffStatus = self["starter"]["starter.current_sensor"].observes
        self["starter"].actuatesProperty = self["speedRatio"]
        self["starter"].electricalOutlet >> self.electricalInlet


VFD_addon_template = {
    "equipment": {("vfd", VFD): {}},
    "properties": {},
}


class PumpWithVFD(Pump):
    """
    This fan is composed of a blower, an electrical motor and a VFD
    """

    _class_iri: URIRef = S223.Pump

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update(
            bases=[pump_template, VFD_addon_template],
            config=config,
        )
        kwargs = {**_config.pop("params", {}), **kwargs}
        _log.debug(f"PumpWithVFD.__init__ {_config} {kwargs}")
        super().__init__(_config, **kwargs)
        self.onOffCommand = self["vfd"]["run_command"]
        self.onOffStatus = self["vfd"]["drive_running"]
        self["vfd"].actuatesProperty = self["speedRatio"]
        self["vfd"].electricalOutlet >> self.electricalInlet



"""
Test Module Template
--------------------
"""

import unittest
import logging

from typing import Callable, Any

_log = logging.getLogger(__name__)


def setup_module() -> None:
    """
    This function is called once at the beginning of all of the tests in this
    module.
    """
    _log.debug("setup_module")


def teardown_module() -> None:
    """
    This function is called once at the end of the tests in this module.
    """
    _log.debug("teardown_module")


def setup_function(function: Callable[..., Any]) -> None:
    """
    This function is called before each module level test function.
    """
    _log.debug("setup_function %r", function)


def teardown_function(function: Callable[..., Any]) -> None:
    """
    This function is called after each module level test function.
    """
    _log.debug("teardown_function %r", function)


def test_some_function(*args: Any, **kwargs: Any) -> None:
    """
    This is a module level test function.
    """
    _log.debug("test_some_function %r %r", args, kwargs)


class TestCaseTemplate(unittest.TestCase):
    @classmethod
    def setup_class(cls) -> None:
        """
        This function is called once before the test case is instantiated
        for each of the tests.
        """
        _log.debug("setup_class")

    @classmethod
    def teardown_class(cls) -> None:
        """
        This function is called once at the end after the last instance
        of the test case has been abandon.
        """
        _log.debug("teardown_class")

    def setup_method(self, method: Callable[..., Any]) -> None:
        """
        This function is called before each test method is called as is
        given a reference to the test method.
        """
        _log.debug("setup_method %r", method)

    def teardown_method(self, method: Callable[..., Any]) -> None:
        """
        This function is called after each test method has been called and
        is given a reference to the test method.
        """
        _log.debug("teardown_method %r", method)

    def test_something(self) -> None:
        """
        This is a method level test function.
        """
        _log.debug("test_something")

    def test_something_else(self) -> None:
        """
        This is another method level test function.
        """
        _log.debug("test_something_else")



"""
Function Blocks
"""

from __future__ import annotations

import inspect
import logging
from datetime import datetime
from typing import Any, AnyStr, Dict

from rdflib import URIRef  # type: ignore

from .core import (
    G36,
    S223,
    Node,
    Property,
    PropertyReference,
    data_graph,
)
from .template import template_update

# logging
_log = logging.getLogger(__name__)

# namespace
_namespace = S223


class FunctionInput(PropertyReference):
    def __new__(cls, arg: Any = None, **kwargs) -> Any:
        _log.debug(f"FunctionInput.__new__ {cls} {arg} {kwargs}")
        if (arg is None) or (arg == ()):
            return Property(**kwargs)
        if isinstance(arg, Property):
            return arg
        if isinstance(arg, (int, float, str, datetime)):
            return Property(arg, **kwargs)

        raise TypeError(f"property expected: {arg}")


class FunctionOutput(PropertyReference):
    def __new__(cls, arg: Any = None, **kwargs) -> Any:
        _log.debug(f"FunctionOutput.__new__ {cls} {arg} {kwargs}")
        if (arg is None) or (arg == ()):
            return Property(**kwargs)
        if isinstance(arg, Property):
            return arg
        if isinstance(arg, (int, float, str, datetime)):
            return Property(arg, **kwargs)

        raise TypeError(f"property expected: {arg}")


#
#   Connector types, parameters, and constants
#


class G36AnalogInput(FunctionInput):
    _class_iri: URIRef = G36.AnalogInput


class G36AnalogOutput(FunctionOutput):
    _class_iri: URIRef = G36.AnalogOutput


class G36DigitalInput(FunctionInput):
    _class_iri: URIRef = G36.DigitalInput


class G36DigitalOutput(FunctionOutput):
    _class_iri: URIRef = G36.DigitalOutput


class Function(Node):
    """
    Function blocks are black boxes representing a sequence or an
    algorithm. Function blocks use inputs and produce outputs that are
    related to observable and actuatable properties.
    Functions are executed by a s223:Contoller.
    """

    _class_iri: URIRef = S223.Function

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update({}, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _log.debug(f"Function.__init__ {kwargs}")

        if not self._resolved:
            self._resolve_annotations()
        _log.debug("    - continue Function.__init__")

        # super().__init__(_config, **kwargs)
        super().__init__(**kwargs)

        # instantiate and associate all of the inputs and outputs
        for attr_name, attr_type in self._nodes.items():
            if not inspect.isclass(attr_type):
                continue
            if not issubclass(attr_type, (FunctionInput, FunctionOutput)):
                continue

            # check if an instance was passed as a kwarg, if it was then
            # the attribute element will be a node and has its _node_iri
            attr_element = getattr(self, attr_name, None)
            if attr_element is None:
                continue

            # if this is used as a function input/output, make it so
            if issubclass(attr_type, FunctionInput):
                data_graph.add((self._node_iri, S223.hasInput, attr_element._node_iri))
                # data_graph.add(
                #     (attr_element._node_iri, RDF.type, S223.FunctionInput)
                # )
            elif issubclass(attr_type, FunctionOutput):
                data_graph.add((self._node_iri, S223.hasOutput, attr_element._node_iri))
                # data_graph.add(
                #     (attr_element._node_iri, RDF.type, S223.FunctionOutput)
                # )

    def __setattr__(self, attr: str, value: Any) -> None:
        """
        .
        """
        # do the normal things
        super().__setattr__(attr, value)

        # if this is a node, double check the type
        attr_type = self._nodes.get(attr, None)
        _log.debug("    - attr_type: %r", attr_type)
        if not attr_type:
            return

        _log.debug(f"Function.__setattr__ {attr} {value}")

        # get the element after it has been set, it will be an instance of
        # attr_type which might not be the value
        attr_element = vars(self).get(attr)
        _log.debug("    - attr_element: %r", attr_element)

        # if this is used as a function input/output, make it so
        if issubclass(attr_type, FunctionInput):
            data_graph.add((self._node_iri, S223.hasInput, attr_element._node_iri))
            # data_graph.add(
            #     (attr_element._node_iri, RDF.type, S223.FunctionInput)
            # )
        elif issubclass(attr_type, FunctionOutput):
            data_graph.add((self._node_iri, S223.hasOutput, attr_element._node_iri))
            # data_graph.add(
            #     (attr_element._node_iri, RDF.type, S223.FunctionOutput)
            # )

    def uses(
        self,
        prop: Property,
        klass: FunctionInput = FunctionInput,
        label: AnyStr = "input",
    ) -> None:
        connector = klass(self, label=f"{self.label}.{label}")
        setattr(self, label, connector)
        prop >> connector

    def produces(
        self,
        prop: Property,
        klass: FunctionOutput = FunctionOutput,
        label: AnyStr = "output",
    ) -> None:
        connector = klass(self, label=f"{self.label}.{label}")
        setattr(self, label, connector)
        connector >> prop



from typing import Dict

from rdflib import URIRef

from ...connections.air import AirInletConnectionPoint, AirOutletConnectionPoint
from ...core import (
    BOB,
    P223,
    Equipment,
    PropertyReference,
)
from ...sensor.gas import (
    COSensor,
    NO2Sensor,
)

# namespace
_namespace = BOB

gasmonitor_template = {
    "params": {
        "label": "Name Of Equipment",
        "comment": "Description",
    },
    "sensors": {
        ("COSensor", COSensor): {
            "hasExternalReference": "bacnet://",
            # "properties": {
            #   "ofSubstance": Constituent.CO,
            #                 ("hasMinRange", QuantifiableObservableProperty): {
            #                     "hasQuantityKind": QUANTITYKIND.DimensionlessRatio,
            #                     "hasUnit": UNIT.PPM,
            #                 },
            #                 ("hasMaxRange", QuantifiableObservableProperty): {
            #                     "hasQuantityKind": QUANTITYKIND.DimensionlessRatio,
            #                     "hasUnit": UNIT.PPM,
            #                 },
            # },
        },
        ("NO2Sensor", NO2Sensor): {
            "hasExternalReference": "bacnet://",
            # "properties": {
            # "ofSubstance": Constituent.NO2,
            #             "hasMinRange": QuantifiableObservableProperty(
            #                 0, hasQuantityKind=QUANTITYKIND.DimensionlessRatio, hasUnit=UNIT.PPM
            #             ),
            #             "hasMaxRange": QuantifiableObservableProperty(
            #                 100, hasQuantityKind=QUANTITYKIND.DimensionlessRatio, hasUnit=UNIT.PPM
            #             ),
            #    },
        },
    },
    "properties": {},
}


class GasMonitor(Equipment):
    """
    Gas monitor that contains 1 or more gas sensors
    """

    _class_iri: URIRef = P223.GasMonitor
    airInletSupply: AirInletConnectionPoint
    airOutletExhaust: AirOutletConnectionPoint

    alarmStatus: PropertyReference

    def __init__(self, config: Dict = gasmonitor_template, **kwargs):
        config["properties"] = config.get(
            "properties", gasmonitor_template["properties"]
        )
        kwargs = {**config.get("params", {}), **kwargs}
        super().__init__(config, **kwargs)



from ..core import BOB, P223
from ..producer import Function, FunctionInput, FunctionOutput

_namespace = BOB


class OccupancyFunction(Function):
    """
    This function takes the the occupancy as detected by a sensor (inStatus,
    perhaps from a motion sensor), the occupancy schedule (inSchedule with
    an external reference to a BACnet Schedule Object) and outputs
    "occupied" if the space/room/zone should be considered occupied.
    """

    _class_iri = P223.OccupancyFunction

    inStatus: FunctionInput
    inOccSensor: FunctionInput
    inLocalOverride: FunctionInput
    inSchedule: FunctionInput
    outStatus: FunctionOutput



from .core import (
    G36,
    P223,
    QUANTITYKIND,
    S223,
    UNIT,
    Constituent,
    Domain,
    EnumerationKind,
    Medium,
    Mix,
    Role,
    Substance,
)

_namespace = S223

# ======================================
# Things that are just Substances
# ======================================
#
Substance.Particle = Particulate = Substance("Particulate")
Particulate.PM1_0 = Particulate.PM1_0 = Particulate(
    "PM1.0"
)  # don't create PM1_0, not clear enough
Particulate.PM2_5 = Particulate.PM2_5 = Particulate("PM2.5")
Particulate.PM10_0 = Particulate.PM10_0 = Particulate("PM10.0")
Substance.Soot = Soot = Substance("Soot")

# ======================================
# Media, Constituents & Mix
# ======================================
#
Constituent.H2O = H2O = Constituent("H2O", label="H2O")
Constituent.Oil = Oil = Constituent("Oil", label="Oil", _alt_namespace=P223)
Constituent.Smoke = Smoke = Constituent("Smoke", label="Smoke", _alt_namespace=P223)
# Gases
Constituent.Ar = Argon = Constituent("Ar", label="Argon", _alt_namespace=P223)
Constituent.CO = CO = Constituent("CO", label="Carbon monoxyde")
Constituent.CO2 = CO2 = Constituent("CO2", label="Carbon dioxyde")
Constituent.NO2 = NO2 = Constituent("NO2", label="NO2", _alt_namespace=P223)
Constituent.CH4 = CH4 = Constituent("CH4", label="CH4", _alt_namespace=P223)
Constituent.NH3 = NH3 = Constituent("NH3", label="NH3", _alt_namespace=P223)
Constituent.H2S = H2S = Constituent("H2S", label="H2S", _alt_namespace=P223)
Constituent.O2 = O2 = Constituent("O2", label="O2", _alt_namespace=P223)
Constituent.O3 = O3 = Constituent("O3", label="O3", _alt_namespace=P223)
Constituent.SO2 = SO2 = Constituent("SO2", label="SO2", _alt_namespace=P223)
Constituent.N = Nitrogen = Constituent("N", label="Nitrogen", _alt_namespace=P223)
Constituent.VOC = VOC = Constituent("VOC", label="VOC", _alt_namespace=P223)
Constituent.Radon = Radon = Constituent("Radon", label="Radon", _alt_namespace=P223)
Constituent.R22 = const_R22 = Constituent("R-22", label="R-22", _alt_namespace=P223)
Constituent.R134A = const_R134A = Constituent(
    "R-134A", label="R-134A", _alt_namespace=P223
)
Constituent.R410A = const_R410A = Constituent(
    "R-410A", label="R-410A", _alt_namespace=P223
)
Constituent.R32 = const_R32 = Constituent("R-32", label="R-32", _alt_namespace=P223)

Constituent.Glycol = Glycol = Constituent("Glycol", label="Glycol")

# Electromagnetic
EM = Constituent("EM")  # electro-magnetic
EM.Light = Light = EM("Light")
EM.Microwave = Microwave = EM("Microwave")
EM.RF = RF = EM("RF")

# Electricity
Electricity = Constituent("Electricity")
Electricity.AC = Electricity("AC")
Electricity.DC = Electricity("DC")

Electricity.AC10000VLL_1Ph_60Hz = Electricity.AC("10000VLL-1Ph-60Hz")
Electricity.AC10000VLL_3Ph_60Hz = Electricity.AC("10000VLL-3Ph-60Hz")
Electricity.AC10000VLL_5770VLN_1Ph_60Hz = Electricity.AC("10000VLL-5770VLN-1Ph-60Hz")
Electricity.AC10000VLL_5770VLN_3Ph_60Hz = Electricity.AC("10000VLL-5770VLN-3Ph-60Hz")
Electricity.AC110VLN_1Ph_50Hz = Electricity.AC("110VLN-1Ph-50Hz")
Electricity.AC120VLN_1Ph_60Hz = Electricity.AC("120VLN-1Ph-60Hz")
Electricity.AC127VLN_1Ph_50Hz = Electricity.AC("127VLN-1Ph-50Hz")
Electricity.AC139VLN_1Ph_50Hz = Electricity.AC("139VLN-1Ph-50Hz")
Electricity.AC1730VLN_1Ph_60Hz = Electricity.AC("1730VLN-1Ph-60Hz")
Electricity.AC1900VLN_1Ph_60Hz = Electricity.AC("1900VLN-1Ph-60Hz")
Electricity.AC190VLL_110VLN_1Ph_50Hz = Electricity.AC("190VLL-110VLN-1Ph-50Hz")
Electricity.AC190VLL_110VLN_3Ph_50Hz = Electricity.AC("190VLL-110VLN-3Ph-50Hz")
Electricity.AC190VLL_1Ph_50Hz = Electricity.AC("190VLL-1Ph-50Hz")
Electricity.AC190VLL_3Ph_50Hz = Electricity.AC("190VLL-3Ph-50Hz")
Electricity.AC208VLL_120VLN_1Ph_60Hz = Electricity.AC("208VLL-120VLN-1Ph-60Hz")
Electricity.AC208VLL_120VLN_3Ph_60Hz = Electricity.AC("208VLL-120VLN-3Ph-60Hz")
Electricity.AC208VLL_1Ph_60Hz = Electricity.AC("208VLL-1Ph-60Hz")
Electricity.AC208VLL_3Ph_60Hz = Electricity.AC("208VLL-3Ph-60Hz")
Electricity.AC219VLN_1Ph_60Hz = Electricity.AC("219VLN-1Ph-60Hz")
Electricity.AC220VLL_127VLN_1Ph_50Hz = Electricity.AC("220VLL-127VLN-1Ph-50Hz")
Electricity.AC220VLL_127VLN_3Ph_50Hz = Electricity.AC("220VLL-127VLN-3Ph-50Hz")
Electricity.AC220VLL_1Ph_50Hz = Electricity.AC("220VLL-1Ph-50Hz")
Electricity.AC220VLL_3Ph_50Hz = Electricity.AC("220VLL-3Ph-50Hz")
Electricity.AC231VLN_1Ph_50Hz = Electricity.AC("231VLN-1Ph-50Hz")
Electricity.AC2400VLN_1Ph_60Hz = Electricity.AC("2400VLN-1Ph-60Hz")
Electricity.AC240VLL_120VLN_1Ph_60Hz = Electricity.AC("240VLL-120VLN-1Ph-60Hz")
Electricity.AC240VLL_139VLN_1Ph_50Hz = Electricity.AC("240VLL-139VLN-1Ph-50Hz")
Electricity.AC240VLL_139VLN_3Ph_50Hz = Electricity.AC("240VLL-139VLN-3Ph-50Hz")
Electricity.AC240VLL_1Ph_50Hz = Electricity.AC("240VLL-1Ph-50Hz")
Electricity.AC240VLL_1Ph_60Hz = Electricity.AC("240VLL-1Ph-60Hz")
Electricity.AC240VLL_208VLN_120VLN_1Ph_60Hz = Electricity.AC(
    "240VLL-208VLN-120VLN-1Ph-60Hz"
)
Electricity.AC240VLL_208VLN_120VLN_3Ph_60Hz = Electricity.AC(
    "240VLL-208VLN-120VLN-3Ph-60Hz"
)
Electricity.AC240VLL_3Ph_50Hz = Electricity.AC("240VLL-3Ph-50Hz")
Electricity.AC240VLL_3Ph_60Hz = Electricity.AC("240VLL-3Ph-60Hz")
Electricity.AC240VLN_1Ph_50Hz = Electricity.AC("240VLN-1Ph-50Hz")
Electricity.AC24VLN_1Ph_50Hz = Electricity.AC("24VLN-1Ph-50Hz")
Electricity.AC24VLN_1Ph_60Hz = Electricity.AC("24VLN-1Ph-60Hz")
Electricity.AC277VLN_1Ph_60Hz = Electricity.AC("277VLN-1Ph-60Hz")
Electricity.AC3000VLL_1730VLN_1Ph_60Hz = Electricity.AC("3000VLL-1730VLN-1Ph-60Hz")
Electricity.AC3000VLL_1730VLN_3Ph_60Hz = Electricity.AC("3000VLL-1730VLN-3Ph-60Hz")
Electricity.AC3000VLL_1Ph_60Hz = Electricity.AC("3000VLL-1Ph-60Hz")
Electricity.AC3000VLL_3Ph_60Hz = Electricity.AC("3000VLL-3Ph-60Hz")
Electricity.AC3300VLL_1900VLN_1Ph_60Hz = Electricity.AC("3300VLL-1900VLN-1Ph-60Hz")
Electricity.AC3300VLL_1900VLN_3Ph_60Hz = Electricity.AC("3300VLL-1900VLN-3Ph-60Hz")
Electricity.AC3300VLL_1Ph_60Hz = Electricity.AC("3300VLL-1Ph-60Hz")
Electricity.AC3300VLL_3Ph_60Hz = Electricity.AC("3300VLL-3Ph-60Hz")
Electricity.AC3460VLN_1Ph_60Hz = Electricity.AC("3460VLN-1Ph-60Hz")
Electricity.AC347VLN_1Ph_60Hz = Electricity.AC("347VLN-1Ph-60Hz")
Electricity.AC380VLL_1Ph_60Hz = Electricity.AC("380VLL-1Ph-60Hz")
Electricity.AC380VLL_219VLN_1Ph_60Hz = Electricity.AC("380VLL-219VLN-1Ph-60Hz")
Electricity.AC380VLL_219VLN_3Ph_60Hz = Electricity.AC("380VLL-219VLN-3Ph-60Hz")
Electricity.AC380VLL_3Ph_60Hz = Electricity.AC("380VLL-3Ph-60Hz")
Electricity.AC3810VLN_1Ph_60Hz = Electricity.AC("3810VLN-1Ph-60Hz")
Electricity.AC400VLL_1Ph_50Hz = Electricity.AC("400VLL-1Ph-50Hz")
Electricity.AC400VLL_231VLN_1Ph_50Hz = Electricity.AC("400VLL-231VLN-1Ph-50Hz")
Electricity.AC400VLL_231VLN_3Ph_50Hz = Electricity.AC("400VLL-231VLN-3Ph-50Hz")
Electricity.AC400VLL_3Ph_50Hz = Electricity.AC("400VLL-3Ph-50Hz")
Electricity.AC415VLL_1Ph_50Hz = Electricity.AC("415VLL-1Ph-50Hz")
Electricity.AC415VLL_240VLN_1Ph_50Hz = Electricity.AC("415VLL-240VLN-1Ph-50Hz")
Electricity.AC415VLL_240VLN_3Ph_50Hz = Electricity.AC("415VLL-240VLN-3Ph-50Hz")
Electricity.AC415VLL_3Ph_50Hz = Electricity.AC("415VLL-3Ph-50Hz")
Electricity.AC4160VLL_1Ph_60Hz = Electricity.AC("4160VLL-1Ph-60Hz")
Electricity.AC4160VLL_2400VLN_1Ph_60Hz = Electricity.AC("4160VLL-2400VLN-1Ph-60Hz")
Electricity.AC4160VLL_2400VLN_3Ph_60Hz = Electricity.AC("4160VLL-2400VLN-3Ph-60Hz")
Electricity.AC4160VLL_3Ph_60Hz = Electricity.AC("4160VLL-3Ph-60Hz")
Electricity.AC480VLL_1Ph_60Hz = Electricity.AC("480VLL-1Ph-60Hz")
Electricity.AC480VLL_277VLN_1Ph_60Hz = Electricity.AC("480VLL-277VLN-1Ph-60Hz")
Electricity.AC480VLL_277VLN_3Ph_60Hz = Electricity.AC("480VLL-277VLN-3Ph-60Hz")
Electricity.AC480VLL_3Ph_60Hz = Electricity.AC("480VLL-3Ph-60Hz")
Electricity.AC5770VLN_1Ph_60Hz = Electricity.AC("5770VLN-1Ph-60Hz")
Electricity.AC6000VLL_1Ph_60Hz = Electricity.AC("6000VLL-1Ph-60Hz")
Electricity.AC6000VLL_3460VLN_1Ph_60Hz = Electricity.AC("6000VLL-3460VLN-1Ph-60Hz")
Electricity.AC6000VLL_3460VLN_3Ph_60Hz = Electricity.AC("6000VLL-3460VLN-3Ph-60Hz")
Electricity.AC6000VLL_3Ph_60Hz = Electricity.AC("6000VLL-3Ph-60Hz")
Electricity.AC600VLL_1Ph_60Hz = Electricity.AC("600VLL-1Ph-60Hz")
Electricity.AC600VLL_347VLN_1Ph_60Hz = Electricity.AC("600VLL-347VLN-1Ph-60Hz")
Electricity.AC600VLL_347VLN_3Ph_60Hz = Electricity.AC("600VLL-347VLN-3Ph-60Hz")
Electricity.AC600VLL_3Ph_60Hz = Electricity.AC("600VLL-3Ph-60Hz")
Electricity.AC6600VLL_1Ph_60Hz = Electricity.AC("6600VLL-1Ph-60Hz")
Electricity.AC6600VLL_3810VLN_1Ph_60Hz = Electricity.AC("6600VLL-3810VLN-1Ph-60Hz")
Electricity.AC6600VLL_3810VLN_3Ph_60Hz = Electricity.AC("6600VLL-3810VLN-3Ph-60Hz")
Electricity.AC6600VLL_3Ph_60Hz = Electricity.AC("6600VLL-3Ph-60Hz")
Electricity.DC12V = Electricity.DC("12V")
Electricity.DC24V = Electricity.DC("24V")
Electricity.DC380V = Electricity.DC("380V")
Electricity.DC48V = Electricity.DC("48V")
Electricity.DC5V = Electricity.DC("5V")
Electricity.DC6V = Electricity.DC("6V")

Electricity.Earth = Electricity("Earth")
Electricity.Neutral = Electricity("Neutral")

Electricity.Signal = Signal = Electricity("Signal")
Electricity.OnOffSignal = Electricity.Signal("OnOffSignal", _alt_namespace=P223)

Signal.RS485 = Signal("EIA485")
Signal.Ethernet = Signal("Ethernet")
Signal.IEC14908 = Signal("IEC14908")
Signal.USB = USB = Signal("USB")
Signal.WiFi = Signal("WiFi", _alt_namespace=P223)
Signal.ModulatedSignal = ModulatedSignal = Signal("Modulated")
ModulatedSignal.DC0_10 = ModulatedSignal("0-10VDC")
ModulatedSignal.MA4_20 = ModulatedSignal("4-20mA")


Mix.PowerAndSignal = PowerAndSignal = Mix("PowerAndSignal", _alt_namespace=P223)
PowerAndSignal.PoE = PowerAndSignal("PoE", _alt_namespace=P223)
PowerAndSignal.PoE.add_constituent(Electricity.DC48V)
PowerAndSignal.PoE.add_constituent(Signal.Ethernet)
PowerAndSignal.USB = PowerAndSignal("USB", _alt_namespace=P223)
PowerAndSignal.USB.add_constituent(Electricity.DC5V)


# ===================
# MEDIA FLAVOURS
# ===================
# Air and gases
Mix.Fluid = Fluid = Mix("Fluid")
Fluid.Air = Air = Fluid("Air")
Fluid.Oil = Oil = Fluid("Oil")
Air.CompressedAir = CompressedAir = Air("CompressedAir")

Fluid.NaturalGas = NaturalGas = Fluid("NaturalGas")
Fluid.Refrigerant = Refrigerant = Fluid("Refrigerant")
Refrigerant.R410A = R410a = Refrigerant("R-410A")
R410a.add_constituent(
    const_R410A, hasQuantityKind=QUANTITYKIND.VolumeFraction, hasUnit=UNIT.PERCENT
)
Refrigerant.R32 = R32 = Refrigerant("R-32", _alt_namespace=P223)
R32.add_constituent(
    const_R32, hasQuantityKind=QUANTITYKIND.VolumeFraction, hasUnit=UNIT.PERCENT
)
Refrigerant.R22 = R22 = Refrigerant("R-22")
R22.add_constituent(
    const_R22, hasQuantityKind=QUANTITYKIND.VolumeFraction, hasUnit=UNIT.PERCENT
)

# Water
Fluid.Water = Water = Fluid("Water")  # constituent in S223 already

Water.ChilledWater = ChilledWater = Water("ChilledWater")

Water.PotableWater = PotableWater = Water("PotableWater", _alt_namespace=P223)

Water.HotWater = HotWater = Water("HotWater")

Water.MixedWater = MixedWater = Water("MixedWater", _alt_namespace=P223)

Water.DomesticWater = DomesticWater = Water("DomesticWater", _alt_namespace=P223)

Water.DomesticHotWater = DomesticHotWater = Water(
    "DomesticHotWater", _alt_namespace=P223
)

Water.Condensate = Condensate = Water("Condensate", _alt_namespace=P223)
Water.Condenser = Condensate = Water("Condenser", _alt_namespace=P223)

Water.GlycolSolution = GlycolSolution = Water(
    "GlycolSolution"
)  # constituent in S223 already

GlycolSolution.GlycolSolution_15Percent = GlycolSolution_15Percent = GlycolSolution(
    "GlycolSolution-15Percent"
)  # constituent in S223 already
GlycolSolution.GlycolSolution_30Percent = GlycolSolution_30Percent = GlycolSolution(
    "GlycolSolution-30Percent"
)  # constituent in S223 already

Water.Steam = Steam = Water("Steam")

# Light
Light.Visible = Light("Visible")
Light.Infrared = Light("Infrared")
Light.Ultraviolet = Light("Ultraviolet")
Occupant = Medium("Occupant")

# ===================
# DOMAINS FLAVOURS
# ===================
Domain.ConveyanceSystems = ConveyanceSystems = Domain("ConveyanceSystems")
Domain.Electrical = Electrical = Domain("Electrical")
Domain.Fire = Fire = Domain("Fire")
Domain.HVAC = HVAC = Domain("HVAC")
Domain.Lighting = Lighting = Domain("Lighting")
Domain.Networking = Networking = Domain("Networking")
Domain.Occupancy = Domain("Occupancy")
Domain.PhysicalSecurity = Security = Domain("PhysicalSecurity")
Domain.Plumbing = Plumbing = Domain("Plumbing")
Domain.Refrigeration = Refrigeration = Domain("Refrigeration")

# ===================
# ===================
Role.Condenser = Condenser = Role("Condenser")
Role.Cooling = Cooling = Role("Cooling")
Role.Discharge = Discharge = Role("Discharge")
Role.Economizer = Economizer = Role("Economizer")
Role.Evaporator = Evaporator = Role("Evaporator")
Role.Exhaust = Exhaust = Role("Exhaust")
Role.Expansion = Expansion = Role("Expansion")
Role.Generator = Generator = Role("Generator")
Role.HeatRecovery = HeatRecovery = Role("HeatRecovery")
Role.HeatTransfer = HeatRecovery = Role("HeatTransfer")
Role.Heating = Heating = Role("Heating")
Role.Load = Load = Role("Load")
Role.OutdoorAirIntake = OutdoorAirIntake = Role("OutdoorAirIntake")
Role.Primary = Primary = Role("Primary")
Role.Recirculating = Recirculating = Role("Recirculating")
Role.Relief = Relief = Role("Relief")
Role.Return = Return = Role("Return")
Role.Secondary = Secondary = Role("Secondary")
Role.Supply = Supply = Role("Supply")

Role.Storage = Storage = Role("Storage", _alt_namespace=P223)

# ===================
# Values Enumeration
# ===================
# Enumeration kinds to create hasValue
ActiveInactiveEnum = EnumerationKind("ActiveInactive", _alt_namespace=P223)
AnalogSignalTypeEnum = EnumerationKind("Analog", _alt_namespace=P223)
BinarySignalTypeEnum = EnumerationKind("BinarySignalType", _alt_namespace=P223)

Binary = EnumerationKind("Binary")
Binary.false = Binary("False")
Binary.true = Binary("True")
#
OnOff = Binary("OnOff")
OnOff.On = OnOff("On")
OnOff.Off = OnOff("Off")
OnOff.Unknown = OnOff("Unknown")

G36AlarmLevel = EnumerationKind("G36AlarmLevels", _alt_namespace=G36)
HandOffAutoEnum = EnumerationKind("HandOffAuto", _alt_namespace=P223)
HVACOperatingMode = EnumerationKind("HVACOperatingMode")
HVACOperatingStatus = EnumerationKind("HVACOperatingStatus")
LeftRightEnum = EnumerationKind("LeftRight", _alt_namespace=P223)
ManualAutoEnum = EnumerationKind("ManualAuto", _alt_namespace=P223)
NiagaraStatusEnum = EnumerationKind("NiagaraStatus", _alt_namespace=P223)  # SEE BELOW
NormalAlarmEnum = EnumerationKind("NormalAlarm", _alt_namespace=P223)
NormalFaultEnum = EnumerationKind("NormalFault", _alt_namespace=P223)
Occupancy = EnumerationKind("Occupancy")
Phase = EnumerationKind("Phase")
Position = EnumerationKind("Position")

OverriddenAuto = EnumerationKind("OverriddenAuto", _alt_namespace=P223)
ProtocolEnum = EnumerationKind("Protocol", _alt_namespace=P223)
RunStatus = EnumerationKind("RunStatus")
Speed = EnumerationKind("Speed")
ThreeSpeedSetting = EnumerationKind("ThreeSpeedSetting")
TopBottomEnum = EnumerationKind("TopBottom", _alt_namespace=P223)
TrueFalseEnum = EnumerationKind("TrueFalse", _alt_namespace=P223)
YesNoEnum = EnumerationKind("YesNo", _alt_namespace=P223)

# Enumerated Values
#
ActiveInactiveEnum.Active = ActiveInactiveEnum("Active", _alt_namespace=P223)
ActiveInactiveEnum.Inactive = ActiveInactiveEnum("Inactive", _alt_namespace=P223)
ActiveInactiveEnum.Unknown = ActiveInactiveEnum("Unknown", _alt_namespace=P223)

#
AnalogSignalTypeEnum.Unknown = AnalogSignalTypeEnum("Unknown", _alt_namespace=P223)
AnalogSignalTypeEnum.RTD = AnalogSignalTypeEnum("RTD", _alt_namespace=P223)
AnalogSignalTypeEnum.Nickel1kRTD = AnalogSignalTypeEnum(
    "Nickel1kRTD", _alt_namespace=P223
)
AnalogSignalTypeEnum.Platinum1kRTD = AnalogSignalTypeEnum(
    "Platinum1kRTD", _alt_namespace=P223
)
AnalogSignalTypeEnum.VDC_0_10 = AnalogSignalTypeEnum("0-10VDC", _alt_namespace=P223)
AnalogSignalTypeEnum.mA_4_20 = AnalogSignalTypeEnum("4-20mA", _alt_namespace=P223)
AnalogSignalTypeEnum.NTC10kType3 = AnalogSignalTypeEnum(
    "NTC10kType3", _alt_namespace=P223
)
AnalogSignalTypeEnum.NTC10kType2 = AnalogSignalTypeEnum(
    "NTC10kType2", _alt_namespace=P223
)
AnalogSignalTypeEnum.NTC2250Type2 = AnalogSignalTypeEnum(
    "NTC2250Type2", _alt_namespace=P223
)
AnalogSignalTypeEnum.Resistive = AnalogSignalTypeEnum("Resistive", _alt_namespace=P223)

#
BinarySignalTypeEnum.DryContact = BinarySignalTypeEnum(
    "DryContact", _alt_namespace=P223
)
BinarySignalTypeEnum.Pulse = BinarySignalTypeEnum("Pulse", _alt_namespace=P223)
BinarySignalTypeEnum.StartStop = BinarySignalTypeEnum("StartStop", _alt_namespace=P223)
BinarySignalTypeEnum.Incremental = BinarySignalTypeEnum(
    "Incremental", _alt_namespace=P223
)

#
G36AlarmLevel.Level1 = EnumerationKind(
    "Level1", comment="Life Safetey Message", _alt_namespace=G36
)
G36AlarmLevel.Level2 = EnumerationKind(
    "Level2", comment="Critical Equipment Message", _alt_namespace=G36
)
G36AlarmLevel.Level3 = EnumerationKind(
    "Level3", comment="Urgent Message", _alt_namespace=G36
)
G36AlarmLevel.Level4 = EnumerationKind(
    "Level4", comment="Normal Message", _alt_namespace=G36
)

#
HandOffAutoEnum.Hand = HandOffAutoEnum("Hand", _alt_namespace=P223)
HandOffAutoEnum.Off = HandOffAutoEnum("Off", _alt_namespace=P223)
HandOffAutoEnum.Auto = HandOffAutoEnum("Auto", _alt_namespace=P223)

#
HVACOperatingMode.Auto = HVACOperatingMode("Auto")
HVACOperatingMode.CoolOnly = HVACOperatingMode("CoolOnly")
HVACOperatingMode.FanOnly = HVACOperatingMode("FanOnly")
HVACOperatingMode.HeatOnly = HVACOperatingMode("HeatOnly")
HVACOperatingMode.Off = HVACOperatingMode("Off")

#
HVACOperatingStatus.Cooling = HVACOperatingStatus("Cooling")
HVACOperatingStatus.Dehumidifying = HVACOperatingStatus("Dehumidifying")
HVACOperatingStatus.Heating = HVACOperatingStatus("Heating")
HVACOperatingStatus.Off = HVACOperatingStatus("Off")
HVACOperatingStatus.Ventilating = HVACOperatingStatus("Ventilating")

#
LeftRightEnum.Left = LeftRightEnum("Left", _alt_namespace=P223)
LeftRightEnum.Right = LeftRightEnum("Right", _alt_namespace=P223)

#
ManualAutoEnum.Manual = ManualAutoEnum("Manual", _alt_namespace=P223)
ManualAutoEnum.Auto = ManualAutoEnum("Auto", _alt_namespace=P223)

# This is a test example... do we want to fill 223 with
# that kind of enums ?
# Or do we provide a way for people to define their own ?
# Niagara is widely used though....
# Same apply to BACnet....
# Alignment ?
NiagaraStatusEnum.ok = NiagaraStatusEnum("ok", _alt_namespace=P223)
NiagaraStatusEnum.unackedAlarm = NiagaraStatusEnum("unackedAlarm", _alt_namespace=P223)
NiagaraStatusEnum.null = NiagaraStatusEnum("null", _alt_namespace=P223)
NiagaraStatusEnum.overridden = NiagaraStatusEnum("overridden", _alt_namespace=P223)
NiagaraStatusEnum.stale = NiagaraStatusEnum("stale", _alt_namespace=P223)
NiagaraStatusEnum.down = NiagaraStatusEnum("down", _alt_namespace=P223)
NiagaraStatusEnum.fault = NiagaraStatusEnum("fault", _alt_namespace=P223)
NiagaraStatusEnum.disabled = NiagaraStatusEnum("disabled", _alt_namespace=P223)
NiagaraStatusEnum.alarm = NiagaraStatusEnum("alarm", _alt_namespace=P223)

#
NormalAlarmEnum.Normal = NormalAlarmEnum("Normal", _alt_namespace=P223)
NormalAlarmEnum.Alarm = NormalAlarmEnum("Alarm", _alt_namespace=P223)

#
NormalFaultEnum.Normal = NormalFaultEnum("Normal", _alt_namespace=P223)
NormalFaultEnum.Fault = NormalFaultEnum("Fault", _alt_namespace=P223)

#
Occupancy.Motion = Motion = Occupancy("Motion")
Occupancy.Occupied = Occupancy("Occupied")
Occupancy.Presence = Presence = Occupancy("Presence")
Presence.false = Presence("False")
Presence.true = Presence("True")
Occupancy.Unknown = Occupancy("Unknown")
Occupancy.Unoccupied = Occupancy("Unoccupied")

#
Motion.true = Motion("True")
Motion.false = Motion("False")

#
Phase.Gas = Phase("Gas")
Phase.Gas.SuperHeated = Phase.Gas("SuperHeated")
Phase.Liquid = Phase("Liquid")
Phase.Liquid.Subcooled = Phase.Liquid("Subcooled")
Phase.Solid = Phase("Solid")
Phase.Vapor = Phase("Vapor")

#
Position.Closed = Position("Closed")
Position.Open = Position("Open")
Position.Unknown = Position("Unknown")

#
OverriddenAuto.Auto = OverriddenAuto("Auto", _alt_namespace=P223)
OverriddenAuto.Overridden = OverriddenAuto("Overridden", _alt_namespace=P223)

#
ProtocolEnum.BACnet = ProtocolEnum("BACnet", _alt_namespace=P223)
ProtocolEnum.BACnet_MSTP = ProtocolEnum("BACnet_MSTP", _alt_namespace=P223)
ProtocolEnum.BACnet_IP = ProtocolEnum("BACnet_IP", _alt_namespace=P223)
ProtocolEnum.BACnet_SC = ProtocolEnum("BACnet_SC", _alt_namespace=P223)
ProtocolEnum.Modbus = ProtocolEnum("Modbus", _alt_namespace=P223)
ProtocolEnum.Modbus_RTU = ProtocolEnum("Modbus_RTU", _alt_namespace=P223)
ProtocolEnum.Modbus_TCP = ProtocolEnum("Modbus_TCP", _alt_namespace=P223)
ProtocolEnum.Lonworks = ProtocolEnum("Lonworks", _alt_namespace=P223)

#
RunStatus.On = RunStatus("On")
RunStatus.Off = RunStatus("Off")
RunStatus.Unknown = RunStatus("Unknown")

#
Speed.High = Speed("High")
Speed.Low = Speed("Low")
Speed.Medium = Speed("Medium")
Speed.Off = Speed("Off")

#
ThreeSpeedSetting.High = ThreeSpeedSetting("High")
ThreeSpeedSetting.Low = ThreeSpeedSetting("Low")
ThreeSpeedSetting.Medium = ThreeSpeedSetting("Medium")
ThreeSpeedSetting.Off = ThreeSpeedSetting("Off")

#
TopBottomEnum.Top = TopBottomEnum("Top", _alt_namespace=P223)
TopBottomEnum.Bottom = TopBottomEnum("Bottom", _alt_namespace=P223)

# lowercase so we don't clash with internal booleans
TrueFalseEnum.true = TrueFalseEnum("True", _alt_namespace=P223)
TrueFalseEnum.false = TrueFalseEnum("False", _alt_namespace=P223)

#
YesNoEnum.Yes = YesNoEnum("Yes", _alt_namespace=P223)
YesNoEnum.No = YesNoEnum("No", _alt_namespace=P223)

# =======
# Aspects
# =======
# Enumeration kinds that add context to properties
Aspect = EnumerationKind("Aspect")
Aspect.Alarm = Aspect("Alarm")
Aspect.CatalogNumber = Aspect("CatalogNumber")
Aspect.Command = Aspect("Command")
Aspect.DayOfWeek = DayOfWeek = Aspect("DayOfWeek")
Aspect.Deadband = Aspect("Deadband")
Aspect.Delta = Aspect("Delta")
Aspect.DryBulb = Aspect("DryBulb")
Aspect.Effectiveness = Effectiveness = Aspect("Effectiveness")
Aspect.Efficiency = Aspect("Efficiency")
Aspect.ElectricalPhaseIdentifier = ElectricalPhaseIdentifier = Aspect(
    "ElectricalPhaseIdentifier"
)
Aspect.ElectricalVoltagePhases = ElectricalVoltagePhases = Aspect(
    "ElectricalVoltagePhases"
)
Aspect.Face = Aspect("Face")
Aspect.Fault = Aspect("Fault")
Aspect.HighLimit = Aspect("HighLimit")
Aspect.Latent = Aspect("Latent")
Aspect.Loss = Aspect("Loss")
Aspect.LowLimit = Aspect("LowLimit")
Aspect.Manufacturer = Aspect("Manufacturer")
Aspect.Maximum = Aspect("Maximum")
Aspect.Minimum = Aspect("Minimum")
Aspect.Model = Aspect("Model")
Aspect.Nominal = Aspect("Nominal")
Aspect.NominalFrequency = Aspect("NominalFrequency")
Aspect.PhaseAngle = Aspect("PhaseAngle")
Aspect.PowerFactor = Aspect("PowerFactor")
Aspect.Rated = Aspect("Rated")
Aspect.Sensible = Aspect("Sensible")
Aspect.SerialNumber = Aspect("SerialNumber")
Aspect.ServiceFactor = Aspect("ServiceFactor")
Aspect.Setpoint = Aspect("Setpoint")
Aspect.StandardConditions = Aspect("StandardConditions")
Aspect.Standby = Aspect("Standby")
Aspect.StartupValue = Aspect("StartupValue")
Aspect.Threshold = Aspect("Threshold")
Aspect.Total = Aspect("Total")
Aspect.WetBulb = Aspect("WetBulb")
Aspect.Year = Aspect("Year")

#
DayOfWeek.Weekday = Weekday = DayOfWeek("Weekday")
DayOfWeek.Weekend = Weekend = DayOfWeek("Weekend")
Weekday.Monday = Weekday("Monday")
Weekday.Tuesday = Weekday("Tuesday")
Weekday.Wednesday = Weekday("Wednesday")
Weekday.Thursday = Weekday("Thursday")
Weekday.Friday = Weekday("Friday")
Weekday.Saturday = Weekday("Saturday")
Weekday.Sunday = Weekday("Sunday")
Weekend.Saturday = Weekday("Saturday")
Weekend.Sunday = Weekday("Sunday")

#
Effectiveness.Active = Effectiveness("Active")
Effectiveness.Inactive = Effectiveness("Inactive", _alt_namespace=P223)
Effectiveness.Unknown = Effectiveness("Unknown", _alt_namespace=P223)

#
Numerical = EnumerationKind("Numerical")
Numerical.DCVoltage = DCVoltage = Numerical("DCVoltage")
DCVoltage.DCNegativeVoltage = DCVoltage("DCNegativeVoltage")
# s223:DCNegativeVoltage-2.5V
# s223:DCNegativeVoltage-3.0V
# s223:DCNegativeVoltage-5.0V
# s223:DCNegativeVoltage-6.0V
# s223:DCNegativeVoltage-12.0V
# s223:DCNegativeVoltage-24.0V
# s223:DCNegativeVoltage-48.0V
# s223:DCNegativeVoltage-190.0V
# s223:DCNegativeVoltage-380.0V
DCVoltage.DCPositiveVoltage = DCVoltage("DCPositiveVoltage")
# see also negative voltages above
DCVoltage.DCZeroVoltage = DCVoltage("DCZeroVoltage")

Numerical.Frequency = Numerical("Frequency")
# s223:Frequency-50Hz
# s223:Frequency-60Hz

Numerical.LineLineVoltage = Numerical("LineLineVoltage")
# many
Numerical.LineNeutralVoltage = Numerical("LineNeutralVoltage")
# many

Numerical.NumberOfElectricalPhases = NumberOfElectricalPhases = Numerical(
    "NumberOfElectricalPhases"
)
NumberOfElectricalPhases.SinglePhase = NumberOfElectricalPhases("SinglePhase")
NumberOfElectricalPhases.ThreePhase = NumberOfElectricalPhases("ThreePhase")

#
Numerical.Voltage = Voltage = Numerical("Voltage")
# many

ElectricalPhaseIdentifier.A = ElectricalPhaseIdentifier("A")
ElectricalPhaseIdentifier.B = ElectricalPhaseIdentifier("B")
ElectricalPhaseIdentifier.C = ElectricalPhaseIdentifier("C")
ElectricalPhaseIdentifier.AB = ElectricalPhaseIdentifier("AB")
ElectricalPhaseIdentifier.BC = ElectricalPhaseIdentifier("BC")
ElectricalPhaseIdentifier.CA = ElectricalPhaseIdentifier("CA")
ElectricalPhaseIdentifier.ABC = ElectricalPhaseIdentifier("ABC")

ElectricalVoltagePhases.ABLineLineVoltage = ElectricalVoltagePhases("ABLineLineVoltage")
ElectricalVoltagePhases.BCLineLineVoltage = ElectricalVoltagePhases("BCLineLineVoltage")
ElectricalVoltagePhases.CALineLineVoltage = ElectricalVoltagePhases("CALineLineVoltage")

ElectricalVoltagePhases.ANLineNeutralVoltage = ElectricalVoltagePhases(
    "ANLineNeutralVoltage"
)
ElectricalVoltagePhases.BNLineNeutralVoltage = ElectricalVoltagePhases(
    "BNLineNeutralVoltage"
)
ElectricalVoltagePhases.CNLineNeutralVoltage = ElectricalVoltagePhases(
    "CNLineNeutralVoltage"
)


## System Types
# SystemType.AirHandlingUnit = AirHandlingUnit = SystemType("AirHandlingUnit")
# SystemType.AirHeatExchanger = AirHeatExchanger = SystemType("AirHeatExchanger")##

# SystemType.Chiller = Chiller = SystemType("Chiller")
# SystemType.CoolingTower = CoolingTower = SystemType("CoolingTower")
# SystemType.FumeHood = FumeHood = SystemType("FumeHood")
# SystemType.Furnace = Furnace = SystemType("Furnace")
# SystemType.HeatExchanger = HeatExchanger = SystemType("HeatExchanger")

# SystemType.HeatPump = HeatPump = SystemType("HeatPump")
# HeatPump.AirToAirHeatPump = AirToAirHeatPump = HeatPump("AirToAirHeatPump")
# HeatPump.GroundToAirHeatPump = GroundToAirHeatPump = HeatPump("GroundToAirHeatPump")
# HeatPump.WaterToAirHeatPump = WaterToAirHeatPump = HeatPump("WaterToAirHeatPump")
# HeatPump.WaterToWaterHeatPump = WaterToWaterHeatPump = HeatPump("WaterToWaterHeatPump")

# SystemType.TerminalUnit = TerminalUnit = SystemType("TerminalUnit")
# TerminalUnit.FanCoilUnit = FanCoilUnit = TerminalUnit("FanCoilUnit")
# TerminalUnit.FanPoweredTerminal = FanPoweredTerminal = TerminalUnit(
#    "FanPoweredTerminal"
# )
# TerminalUnit.SingleDuctTerminal = SingleDuctTerminal = TerminalUnit(
#    "SingleDuctTerminal"
# )
# TerminalUnit.DualDuctTerminal = DualDuctTerminal = TerminalUnit("DualDuctTerminal")

# SystemType.WaterHeater = WaterHeater = SystemType("WaterHeater")



from typing import Any

from bob.properties.states import OnOffStatus

from ..core import (
    S223,
    PropertyReference,
)
from .sensor import Sensor, split_kwargs

_namespace = S223


class IntrusionSensor(Sensor):
    _class_iri = S223.Sensor
    observes: PropertyReference  # OnOffStatus

    def __init__(self, **kwargs: Any) -> None:
        _sensor_kwargs, _property_kwargs = split_kwargs(kwargs)

        super().__init__(**_sensor_kwargs)
        self.observes = OnOffStatus(
            label=f"{self.label}.intrusion_status", **_property_kwargs
        )



from typing import Any


from ..connections.controlsignal import OnOffSignalOutletConnectionPoint
from ..core import (
    BOB,
    S223,
    PropertyReference,
)
from ..enum import Air
from ..properties import SmokePresence
from .sensor import Sensor, split_kwargs

_namespace = BOB


class SmokeDetectionSensor(Sensor):
    _class_iri = S223.Sensor
    observes: PropertyReference  # Temperature
    dryContactOutlet: OnOffSignalOutletConnectionPoint

    def __init__(self, **kwargs: Any) -> None:
        _sensor_kwargs, _property_kwargs = split_kwargs(kwargs)

        if "hasUnit" not in _property_kwargs:
            raise ValueError(
                "You must provide hasUnit when defining a smoke detection sensor"
            )
        if "ofMedium" not in _property_kwargs:
            raise ValueError(
                "You must provide ofMedium when defining a smoke detection sensor"
            )

        super().__init__(**_sensor_kwargs)

        self.observes = SmokePresence(
            # isObservedBy=self,
            label=f"{self.label}.SmokeDetection",
            ofMedium=Air,
            **_property_kwargs,
        )



from bob.enum import AnalogSignalTypeEnum, BinarySignalTypeEnum

from ..core import (
    BOB,
    P223,
    S223,
    Connection,
    ConnectionPoint,
    InletConnectionPoint,
    OutletConnectionPoint,
)
from ..enum import Electricity, ModulatedSignal

_namespace = BOB
# It is modeling dry contact, Triac and other On-Off relationships


# === GENERAL
class OnOffSignalConnection(Connection):
    hasMedium = Electricity.OnOffSignal
    _class_iri = S223.Connection


class OnOffSignalConnectionPoint(ConnectionPoint):
    _attr_uriref = {"hasSignalType": P223.hasSignalType}

    hasMedium = Electricity.OnOffSignal
    hasSignalType: BinarySignalTypeEnum


class OnOffSignalInletConnectionPoint(InletConnectionPoint, OnOffSignalConnectionPoint):
    _class_iri = P223.BinaryInput


class OnOffSignalOutletConnectionPoint(
    OutletConnectionPoint, OnOffSignalConnectionPoint
):
    _class_iri = P223.BinaryOutput


# This is high level and we don't know if it's using 0-10VDC, 4-20mA, etc...
# === Modulation signals


class ModulationSignalConnection(Connection):
    hasMedium = ModulatedSignal
    _class_iri = S223.Connection


class ModulationSignalConnectionPoint(ConnectionPoint):
    _attr_uriref = {"hasSignalType": P223.hasSignalType}

    hasMedium = ModulatedSignal
    hasSignalType: AnalogSignalTypeEnum


class ModulationSignalInletConnectionPoint(
    InletConnectionPoint, ModulationSignalConnectionPoint
):
    _class_iri = P223.AnalogInput


class ModulationSignalOutletConnectionPoint(
    OutletConnectionPoint, ModulationSignalConnectionPoint
):
    _class_iri = P223.AnalogOutput



from __future__ import annotations

from typing import Any

from rdflib import URIRef

from ..core import (
    BOB,
    QUANTITYKIND,
    S223,
    UNIT,
    Constituent,
    PropertyReference,
    Setpoint,
)
from ..properties import GasConcentration
from .sensor import Sensor, split_kwargs

_namespace = BOB

# TODO :
# try to create an exmaple for the sensors found here
# Sal will like :0)
# And those are from Quebec
# http://operadetectors.com/category/gas-monitors-1.aspx


class GasConcentrationSetpoint(Setpoint):
    _class_iri = S223.Setpoint
    hasQuantityKind: URIRef = QUANTITYKIND.DimensionlessRatio
    hasUnit: URIRef = UNIT.PPM


class GasConcentrationSensor(Sensor):
    _class_iri = S223.Sensor
    observes: PropertyReference  # GasConcentration
    hasMinRange: PropertyReference
    hasMaxRange: PropertyReference

    def __init__(self, **kwargs: Any) -> None:
        _sensor_kwargs, _property_kwargs = split_kwargs(kwargs)
        print(f"{_sensor_kwargs = }, {_property_kwargs = }")

        if "ofSubstance" not in _property_kwargs:
            raise ValueError(
                "You must provide ofSubstance when defining a gas concentration sensor"
            )

        super().__init__(**_sensor_kwargs)

        self.observes = GasConcentration(
            # isObservedBy=self,
            label=f"{self.label}.GasConcentration",  # needs more focus
            **_property_kwargs,
        )


class CO2Sensor(GasConcentrationSensor):
    _class_iri = S223.Sensor
    "Carbon Dioxide concentration sensor"
    hasMinRange: PropertyReference
    hasMaxRange: PropertyReference

    def __init__(self, **kwargs):
        # _sensor_kwargs, _property_kwargs = split_kwargs(kwargs)
        super().__init__(ofSubstance=Constituent.CO2, **kwargs)


class COSensor(GasConcentrationSensor):
    _class_iri = S223.Sensor
    "Carbon monoxide concentration sensor"
    hasMinRange: PropertyReference
    hasMaxRange: PropertyReference

    def __init__(self, **kwargs):
        # _sensor_kwargs, _property_kwargs = split_kwargs(kwargs)
        super().__init__(ofSubstance=Constituent.CO, **kwargs)


class NO2Sensor(GasConcentrationSensor):
    _class_iri = S223.Sensor
    "Diesel (NO2) concentration sensor"
    hasMinRange: PropertyReference
    hasMaxRange: PropertyReference

    def __init__(self, **kwargs):
        # _sensor_kwargs, _property_kwargs = split_kwargs(kwargs)
        super().__init__(ofSubstance=Constituent.NO2, **kwargs)


class CH4Sensor(GasConcentrationSensor):
    _class_iri = S223.Sensor
    "Natural gas sensor"
    hasMinRange: PropertyReference
    hasMaxRange: PropertyReference

    def __init__(self, **kwargs):
        # _sensor_kwargs, _property_kwargs = split_kwargs(kwargs)
        super().__init__(ofSubstance=Constituent.CH4, **kwargs)



from typing import Dict

from rdflib import URIRef

from ...connections.liquid import (
    WaterInletConnectionPoint,
    WaterOutletConnectionPoint,
)
from ...core import BOB, P223, Equipment
from ...template import configure_relations, template_update

_namespace = BOB

"""
chilledWaterCoil_template = {
    "params": {"label": "Name", "comment": "Description"},
    "sensors": {},
    "equipment": {("valve", Equipment): {"comment": "SubDev comment"}},
}
"""
# SEMANTIC QUESTION
# here, that could be a good way to define the coil and its valve...
# but the valve connect to the coil
# can this be considered "contained" in the Coil Equipment ?
# Should this b ea system


class GeothermalWell(Equipment):
    _class_iri: URIRef = P223.GeothermalWell
    waterInlet: WaterInletConnectionPoint
    waterOutlet: WaterOutletConnectionPoint

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update({}, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _relations = _config.pop("relations", [])
        super().__init__(_config, **kwargs)
        configure_relations(self, _relations)
        self.waterOutlet.paired_to(self.waterInlet)



import logging
from typing import Dict

from ...core import BOB, S223, Equipment, Property
from ...template import configure_relations, template_update

# logging
_log = logging.getLogger(__name__)

_namespace = BOB

"""
 
__|___|__|___|___|__                                                                                  
|     Actuator     |------------s223:hasProperty--------(actuates) -> A                                   
|  s223:Equipment  |------------s223:hasProperty--------(commandedByProperty) -> B
|                  |
|                  |
|__________________|

"""


class Actuator(Equipment):
    _class_iri = S223.Actuator
    actuates: Equipment

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update({}, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _log.info(f"Actuator.__init__ {_config} {kwargs}")
        _relations = _config.pop("relations", [])
        super().__init__(_config, **kwargs)
        configure_relations(self, _relations)

    def actuatedby(self, actuatedby: Property):
        self._data_graph.add(
            (self._node_iri, S223.actuatedByProperty, actuatedby._node_iri)
        )



from typing import Any

from rdflib import URIRef


from ..core import (
    BOB,
    QUANTITYKIND,
    S223,
    PropertyReference,
    Setpoint,
)
from ..enum import Air, Water
from ..properties import Flow
from .sensor import Sensor, split_kwargs

_namespace = BOB


class FlowSetpoint(Setpoint):
    _class_iri = S223.Setpoint
    hasQuantityKind: URIRef = QUANTITYKIND.VolumeFlowRate
    hasUnit: URIRef


class FlowSensor(Sensor):
    _class_iri = S223.Sensor
    observes: PropertyReference  # Flow

    def __init__(self, **kwargs: Any) -> None:
        _sensor_kwargs, _property_kwargs = split_kwargs(kwargs)

        super().__init__(**_sensor_kwargs)

        self.observes = Flow(
            # isObservedBy=self,
            label=f"{self.label}.Flow",
            **_property_kwargs,
        )


class AirFlowSensor(FlowSensor):
    _class_iri = S223.Sensor

    # typical unit : hasUnit=UNIT["FT3-PER-MIN"]
    def __init__(self, **kwargs):
        super().__init__(ofMedium=Air, **kwargs)


class WaterFlowSensor(FlowSensor):
    _class_iri = S223.Sensor

    # typical unit : hasUnit=UNIT["GAL_UK-PER-MIN"]
    def __init__(self, **kwargs):
        super().__init__(ofMedium=Water, **kwargs)



from ..core import BOB, QUANTITYKIND, UNIT, Medium, QuantifiableObservableProperty

_namespace = BOB


class Gallons(QuantifiableObservableProperty):
    hasQuantityKind = QUANTITYKIND.LiquidVolume
    hasUnit = UNIT.GAL_US
    measuresMedium: Medium  # set from the sensor



from ..core import (
    BOB,
    QUANTITYKIND,
    UNIT,
    Medium,
    QuantifiableActuatableProperty,
    QuantifiableObservableProperty,
)
from ..enum import Air

_namespace = BOB


class Percent(QuantifiableObservableProperty):
    hasQuantityKind = QUANTITYKIND.DimensionlessRatio
    hasUnit = UNIT.PERCENT


class PercentCommand(QuantifiableActuatableProperty):
    hasQuantityKind = QUANTITYKIND.DimensionlessRatio
    hasUnit = UNIT.PERCENT


class RPM(QuantifiableObservableProperty):
    hasQuantityKind = QUANTITYKIND.AngularVelocity
    hasUnit = UNIT["REV-PER-MIN"]


class RelativeHumidity(QuantifiableObservableProperty):
    hasQuantityKind = QUANTITYKIND.RelativeHumidity
    hasUnit = UNIT.PERCENT_RH
    ofMedium: Medium = Air


class GasConcentration(QuantifiableObservableProperty):
    hasQuantityKind = QUANTITYKIND.DimensionlessRatio
    hasUnit = UNIT.PPM
    ofMedium: Medium = Air



from bob.connections.naturalgas import NaturalGasInletConnectionPoint
from bob.properties import PercentCommand

from ...connections.air import AirInletConnectionPoint, AirOutletConnectionPoint
from ...connections.electricity import ElectricalInletConnectionPoint
from ...connections.liquid import (
    SteamInletConnectionPoint,
    SteamOutletConnectionPoint,
    WaterInletConnectionPoint,
)
from ...core import BOB, S223, Equipment

_namespace = BOB


class SteamPipe(Equipment):
    # One way of providing humidity to air
    # a simple pie with steam
    airInlet: AirInletConnectionPoint
    airOutlet: AirOutletConnectionPoint
    steamInlet: SteamInletConnectionPoint  # from the humidifier


class Humidifier(Equipment):
    _class_iri = S223.Humidifier
    steamOutlet: SteamOutletConnectionPoint
    waterInlet: WaterInletConnectionPoint


class ElectricalHumidifier(Humidifier):
    _class_iri = S223.Humidifier
    powerInlet: ElectricalInletConnectionPoint
    modulation = PercentCommand


class NaturalGasHumidifier(Humidifier):
    _class_iri = S223.Humidifier
    naturalGasInlet: NaturalGasInletConnectionPoint
    modulation = PercentCommand



from rdflib import URIRef

from ..core import (
    BOB,
    S223,
    EnumerableProperty,
    EnumeratedActuatableProperty,
    EnumeratedObservableProperty,
    Medium,
    Substance,
)
from ..enum import (
    Light,
    Motion,
    NormalAlarmEnum,
    Occupancy,
    OnOff,
    Position,
    Smoke,
    YesNoEnum,
)

_namespace = BOB


# On Off Status is telemetry so the value depends on hasExternalReference
# Using EnumerationKind will lead to the reading being written down in the model
# which is not what we want. We want to know where to find this real time status.


class OnOffStatus(EnumeratedObservableProperty):
    _class_iri: URIRef = S223.EnumeratedObservableProperty
    hasEnumerationKind = OnOff
    hasValue: OnOff


class OnOffCommand(EnumeratedActuatableProperty):
    _class_iri: URIRef = S223.EnumeratedActuatableProperty
    hasEnumerationKind = OnOff
    hasValue: OnOff


class NormalAlarmStatus(EnumeratedObservableProperty):
    _class_iri: URIRef = S223.EnumeratedObservableProperty
    hasEnumerationKind = NormalAlarmEnum
    hasValue: NormalAlarmEnum


class OpenCloseCommand(EnumeratedActuatableProperty):
    _class_iri: URIRef = S223.EnumeratedActuatableProperty
    hasEnumerationKind = Position
    hasValue: Position


class OpenCloseStatus(EnumeratedActuatableProperty):
    _class_iri: URIRef = S223.EnumeratedActuatableProperty
    hasEnumerationKind = Position
    hasValue: Position


class Schedule(EnumerableProperty):
    _class_iri: URIRef = S223.EnumerableProperty
    hasEnumerationKind = Occupancy
    hasValue: Occupancy


class OccupancyStatus(EnumeratedObservableProperty):
    _class_iri: URIRef = S223.EnumeratedObservableProperty
    hasEnumerationKind = Occupancy
    hasValue: Occupancy


class Motion(EnumeratedObservableProperty):
    _class_iri: URIRef = S223.EnumeratedObservableProperty
    hasEnumerationKind = Motion
    hasValue: Motion


class SmokePresence(EnumeratedObservableProperty):
    ofMedium: Medium  # set from the sensor
    ofSubstance: Substance = Smoke
    # isObservedBy: Sensor
    hasEnumerationKind = YesNoEnum
    hasValue: YesNoEnum


class DaylightDetected(EnumeratedObservableProperty):
    ofMedium: Light.Visible
    hasEnumerationKind = YesNoEnum
    hasValue: YesNoEnum



from typing import Dict

from rdflib import URIRef

from bob.core import BOB, P223, S223, UNIT, Equipment, PropertyReference
from bob.properties.states import OnOffStatus

from ...connections.controlsignal import (
    ModulationSignalOutletConnectionPoint,
    OnOffSignalInletConnectionPoint,
    OnOffSignalOutletConnectionPoint,
)
from ...connections.network import (
    RS485BidirectionalConnectionPoint,
)
from ...sensor.humidity import AirHumiditySensor
from ...sensor.pressure import AirDifferentialStaticPressureSensor
from ...sensor.temperature import AirTemperatureSensor, TemperatureSetpoint

_namespace = BOB


class Thermostat(Equipment):
    _class_iri = S223.Thermostat
    temperature: PropertyReference
    setpoint: PropertyReference
    differential: PropertyReference
    onOffCommand: PropertyReference


class Pressurestat(Equipment):
    _class_iri = P223.Pressurestat
    pressure: PropertyReference
    setpoint: PropertyReference
    differential: PropertyReference
    onOffCommand: PropertyReference


class Humidistat(Equipment):
    _class_iri = S223.Humidistat
    humidity: PropertyReference
    setpoint: PropertyReference
    differential: PropertyReference
    onOffCommand: PropertyReference


MechanicalOnOffThermostat_template = {
    "cp": {
        "mstp": RS485BidirectionalConnectionPoint,
        "heatingOutput": OnOffSignalOutletConnectionPoint,
        "coolingOutput": OnOffSignalOutletConnectionPoint,
        "fanOutput": OnOffSignalOutletConnectionPoint,
    },
    "properties": {
        ("temperature_setpoint", TemperatureSetpoint): {"hasUnit": UNIT.DEG_C}
    },
    "sensors": {
        ("temperature_sensor", AirTemperatureSensor): {"hasUnit": UNIT.DEG_C},
        ("humidity_sensor", AirHumiditySensor): {},
    },
}


class MechanicalOnOffThermostat(Thermostat):
    """
    A mechanical thermostat have no network connection
    Outputs are turned on and off depending on the setpoint
    and the temperature read by the sensor inside the thermostat
    """

    def __init__(self, config: Dict = MechanicalOnOffThermostat_template, **kwargs):
        config["properties"] = config.get(
            "properties", MechanicalOnOffThermostat_template["properties"]
        )
        kwargs = {**config.get("params", {}), **kwargs}
        super().__init__(config, **kwargs)
        self.temperature = self["temperature_sensor"].observedProperty


MechanicalModulatingThermostat_template = {
    "cp": {
        "mstp": RS485BidirectionalConnectionPoint,
        "heatingOutput": ModulationSignalOutletConnectionPoint,
        "coolingOutput": ModulationSignalOutletConnectionPoint,
        "fanOutput": OnOffSignalOutletConnectionPoint,
    },
    "properties": {
        ("temperature_setpoint", TemperatureSetpoint): {"hasUnit": UNIT.DEG_C}
    },
    "sensors": {
        ("temperature_sensor", AirTemperatureSensor): {"hasUnit": UNIT.DEG_C},
        ("humidity_sensor", AirHumiditySensor): {},
    },
}


class MechanicalModulatingThermostat(Thermostat):
    """
    A mechanical thermostat have no network connection
    This model is modulating so all output are modulation signals
    Outputs are controller depending on the setpoint
    and the temperature read by the sensor inside the thermostat
    """

    def __init__(
        self, config: Dict = MechanicalModulatingThermostat_template, **kwargs
    ):
        config["properties"] = config.get(
            "properties", MechanicalModulatingThermostat_template["properties"]
        )
        kwargs = {**config.get("params", {}), **kwargs}
        super().__init__(config, **kwargs)


NetworkThermostat_template = {
    "cp": {
        "mstp": RS485BidirectionalConnectionPoint,
        "heatingOutput": OnOffSignalOutletConnectionPoint,
        "coolingOutput": OnOffSignalOutletConnectionPoint,
    },
    "properties": {
        ("temperature_setpoint", TemperatureSetpoint): {"hasUnit": UNIT.DEG_C}
    },
    "sensors": {
        ("temperature_sensor", AirTemperatureSensor): {"hasUnit": UNIT.DEG_C},
        ("humidity_sensor", AirHumiditySensor): {},
    },
}


class NetworkThermostat(Thermostat):
    """
    A network thermostat has the ability to control loads
    direclty from outputs.
    ex. TEC3000
    """

    def __init__(self, config: Dict = NetworkThermostat_template, **kwargs):
        config["properties"] = config.get(
            "properties", NetworkThermostat_template["properties"]
        )
        kwargs = {**config.get("params", {}), **kwargs}
        super().__init__(config, **kwargs)


NetworkRoomSensor_template = {
    "cp": {"mstp": RS485BidirectionalConnectionPoint},
    "properties": {
        ("temperature_setpoint", TemperatureSetpoint): {"hasUnit": UNIT.DEG_C}
    },
    "sensors": {
        ("temperature_sensor", AirTemperatureSensor): {"hasUnit": UNIT.DEG_C},
        ("humidity_sensor", AirHumiditySensor): {},
    },
}


class NetworkRoomSensor(Equipment):
    """
    A network Room Sensor tells information on the room and accept setpoints
    Will communicate thoses information by network.
    But no outputs to activate loads.
    """

    _class_iri: URIRef = P223.NetworkRoomSensor

    def __init__(self, config: Dict = NetworkRoomSensor_template, **kwargs):
        config["properties"] = config.get(
            "properties", NetworkRoomSensor_template["properties"]
        )
        kwargs = {**config.get("params", {}), **kwargs}
        super().__init__(config, **kwargs)


HighStaticPressureStat_template = {
    "cp": {
        "resetInput": OnOffSignalInletConnectionPoint,
        "signalOutput": OnOffSignalOutletConnectionPoint,
    },
    "properties": {("onOffStatus", OnOffStatus): {}},
    "sensors": {
        ("pressure_sensor", AirDifferentialStaticPressureSensor): {"hasUnit": UNIT.PA}
    },
}


# Pressure
class HighStaticPressureStat(Pressurestat):
    def __init__(self, config: Dict = HighStaticPressureStat_template, **kwargs):
        config["properties"] = config.get(
            "properties", HighStaticPressureStat_template["properties"]
        )
        kwargs = {**config.get("params", {}), **kwargs}
        super().__init__(config, **kwargs)


flowswitch_template = {
    "cp": {"signalOutput": OnOffSignalOutletConnectionPoint},
    "properties": {("onOffStatus", OnOffStatus): {}},
    "sensors": {
        ("pressure_sensor", AirDifferentialStaticPressureSensor): {"hasUnit": UNIT.PA}
    },
}


class FlowSwitch(Equipment):
    """
    A contact On Off controlled by static pressure in duct
    """

    _class_iri: URIRef = P223.Flowswitch

    def __init__(self, config: Dict = flowswitch_template, **kwargs):
        config["properties"] = config.get(
            "properties", flowswitch_template["properties"]
        )
        kwargs = {**config.get("params", {}), **kwargs}
        super().__init__(config, **kwargs)



from typing import Dict

from ...connections.liquid import (
    FluidBidirectionalConnectionPoint,
    FluidInletConnectionPoint,
    FluidOutletConnectionPoint,
)
from ...core import BOB, P223, Equipment
from ...enum import Fluid, Water
from ...properties.force import Pressure
from ...properties.temperature import Temperature
from ...template import template_update

_namespace = BOB


class Tank(Equipment):
    _class_iri = P223.Tank
    fluidInlet: FluidInletConnectionPoint
    fluidOutlet: FluidOutletConnectionPoint
    containedFluid: FluidBidirectionalConnectionPoint

    fluidTemperature: Temperature
    internalPressure: Pressure

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update({}, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        super().__init__(_config, **kwargs)
        self.fluidOutlet.paired_to(self.fluidInlet)

    def set_fluid_type(self, fluid: Fluid = Water):
        self.set_medium(["leavingFluid", "enteringFluid", "containedFluid"], fluid)



from rdflib import URIRef

from .core import P223, Node

_namespace = P223


class AirHandlingUnit(Node):
    _class_iri: URIRef = P223["Application-AirHandlingUnit"]


class Boiler(Node):
    _class_iri: URIRef = P223["Application-Boiler"]


class Chiller(Node):
    _class_iri: URIRef = P223["Application-Chiller"]


class CoolingTower(Node):
    _class_iri: URIRef = P223["Application-CoolingTower"]


class FumeHood(Node):
    _class_iri: URIRef = P223["Application-FumeHood"]


class Furnace(Node):
    _class_iri: URIRef = P223["Application-Furnace"]


class HeatExchanger(Node):
    _class_iri: URIRef = P223["Application-HeatExchanger"]


class HeatPump(Node):
    _class_iri: URIRef = P223["Application-HeatPump"]


class HotWaterHeater(Node):
    _class_iri: URIRef = P223["Application-HotWaterHeater"]


class AirToAirHeatPump(HeatPump):
    _class_iri: URIRef = P223["HeatPump-AirToAirHeatPump"]


class GroundToAirHeatPump(HeatPump):
    _class_iri: URIRef = P223["HeatPump-GroundToAirHeatPump"]


class WaterToAirHeatPump(HeatPump):
    _class_iri: URIRef = P223["HeatPump-WaterToAirHeatPump"]


class WaterToWaterHeatPump(HeatPump):
    _class_iri: URIRef = P223["HeatPump-WaterToWaterHeatPump"]


class TerminalUnit(Node):
    _class_iri: URIRef = P223["Application-TerminalUnit"]


class FanCoilUnit(TerminalUnit):
    _class_iri: URIRef = P223["TerminalUnitApplication-FanCoilUnit"]


class FanPoweredTerminal(TerminalUnit):
    _class_iri: URIRef = P223["TerminalUnitApplication-FanPoweredTerminal"]


class SingleDuctTerminal(TerminalUnit):
    _class_iri: URIRef = P223["TerminalUnitApplication-SingleDuctTerminal"]


class DualDuctTerminal(TerminalUnit):
    _class_iri: URIRef = P223["TerminalUnitApplication-DualDuctTerminal"]


class ElectricalDistribution(Node):
    _class_iri: URIRef = P223["Application-ElectricalDistribution"]


class ElectricalPanel(ElectricalDistribution):
    _class_iri: URIRef = P223["ElectricalDistribution-ElectricalPanel"]



from typing import Any

from rdflib import URIRef

from ..core import (
    BOB,
    P223,
    QUANTITYKIND,
    UNIT,
    QuantifiableObservableProperty,
)

_namespace = BOB


class Volts(QuantifiableObservableProperty):
    _node_iri = P223.Volts
    hasQuantityKind = QUANTITYKIND.Voltage
    hasUnit = UNIT.V


class Amps(QuantifiableObservableProperty):
    _node_iri = P223.Amps
    hasQuantityKind = QUANTITYKIND.ElectricCurrent
    hasUnit = UNIT.A


class PowerFactor(QuantifiableObservableProperty):
    _node_iri = P223.PowerFactor
    hasQuantityKind = QUANTITYKIND.PowerFactor
    hasUnit = UNIT.UNITLESS


class Frequency(QuantifiableObservableProperty):
    _node_iri = P223.Frequency
    hasQuantityKind = QUANTITYKIND.Frequency
    hasUnit = UNIT.HZ


class ElectricPower(QuantifiableObservableProperty):
    _node_iri = P223.ElectricPower
    hasQuantityKind = QUANTITYKIND.ElectricPower
    hasUnit: URIRef
    _supported_units = [
        UNIT.KiloW,
        UNIT.W,
        UNIT.TeraW,
        UNIT.PicoW,
        UNIT.NanoW,
        UNIT.MilliW,
        UNIT.MicroW,
        UNIT.MegaW,
        UNIT.HP_Electric,
        UNIT.GigaW,
    ]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if not ("hasUnit" in kwargs and kwargs["hasUnit"] in self._supported_units):
            _unit = kwargs["hasUnit"] if "hasUnit" in kwargs else "None"
            raise ValueError(
                f"You must provide hasUnit when defining {self}. This unit must be one of those types : {self._supported_units}. You provided {_unit}"
            )
        super().__init__(*args, **kwargs)


class ElectricApparentPower(QuantifiableObservableProperty):
    _node_iri = P223.ElectricApparentPower
    hasQuantityKind = QUANTITYKIND.ApparentPower
    hasUnit: URIRef
    _supported_units = [UNIT["V-A"], UNIT["KiloV-A"]]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if not ("hasUnit" in kwargs and kwargs["hasUnit"] in self._supported_units):
            _unit = kwargs["hasUnit"] if "hasUnit" in kwargs else "None"
            raise ValueError(
                f"You must provide hasUnit when defining {self}. This unit must be one of those types : {self._supported_units}. You provided {_unit}"
            )
        super().__init__(*args, **kwargs)


class ElectricReactivePower(QuantifiableObservableProperty):
    _node_iri = P223.ElectricReactivePower
    hasQuantityKind = QUANTITYKIND.ReactivePower
    hasUnit: URIRef
    _supported_units = [UNIT["V-A_Reactive"], UNIT["KiloV-A_Reactive"]]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if not ("hasUnit" in kwargs and kwargs["hasUnit"] in self._supported_units):
            _unit = kwargs["hasUnit"] if "hasUnit" in kwargs else "None"
            raise ValueError(
                f"You must provide hasUnit when defining {self}. This unit must be one of those types : {self._supported_units}. You provided {_unit}"
            )
        super().__init__(*args, **kwargs)


# Super common....so making a shortcut


class ElectricPowerkW(ElectricPower):
    def __init__(self, **kwargs):
        kwargs["hasUnit"] = UNIT.KiloW
        super().__init__(**kwargs)


class ElectricPowerW(ElectricPower):
    def __init__(self, **kwargs):
        kwargs["hasUnit"] = UNIT.W
        super().__init__(**kwargs)


# Energy
class ElectricEnergy(QuantifiableObservableProperty):
    _node_iri = P223.ElectricEnergy
    hasQuantityKind = QUANTITYKIND.Energy
    hasUnit: URIRef
    _supported_units = [
        UNIT["KiloW-HR"],
        UNIT["W-HR"],
        UNIT["TeraW-HR"],
        UNIT["MegaW-HR"],
        UNIT["GigaW-HR"],
    ]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if not ("hasUnit" in kwargs and kwargs["hasUnit"] in self._supported_units):
            _unit = kwargs["hasUnit"] if "hasUnit" in kwargs else "None"
            raise ValueError(
                f"You must provide hasUnit when defining {self}. This unit must be one of those types : {self._supported_units}. You provided {_unit}"
            )
        super().__init__(*args, **kwargs)


class ElectricApparentEnergy(QuantifiableObservableProperty):
    _node_iri = P223.ElectricApparentEnergy
    hasQuantityKind = QUANTITYKIND.Energy
    hasUnit: URIRef
    _supported_units = [UNIT["V-A-HR"], UNIT["KiloV-A-HR"]]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if not ("hasUnit" in kwargs and kwargs["hasUnit"] in self._supported_units):
            _unit = kwargs["hasUnit"] if "hasUnit" in kwargs else "None"
            raise ValueError(
                f"You must provide hasUnit when defining {self}. This unit must be one of those types : {self._supported_units}. You provided {_unit}"
            )
        super().__init__(*args, **kwargs)


class ElectricReactiveEnergy(QuantifiableObservableProperty):
    _node_iri = P223.ElectricReactiveEnergy
    hasQuantityKind = QUANTITYKIND.Energy
    hasUnit: URIRef
    _supported_units = [UNIT["V-A_Reactive-HR"], UNIT["KiloV-A_Reactive-HR"]]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if not ("hasUnit" in kwargs and kwargs["hasUnit"] in self._supported_units):
            _unit = kwargs["hasUnit"] if "hasUnit" in kwargs else "None"
            raise ValueError(
                f"You must provide hasUnit when defining {self}. This unit must be one of those types : {self._supported_units}. You provided {_unit}"
            )
        super().__init__(*args, **kwargs)



from typing import Dict


from bob.properties import PercentCommand
from bob.properties.electricity import Amps, ElectricPowerkW
from bob.properties.states import OnOffCommand

from ...connections.air import (
    AirBidirectionalConnectionPoint,
    AirInletConnectionPoint,
    AirOutletConnectionPoint,
)
from ...connections.electricity import (
    Electricity_240VLL_1Ph_60HzInletConnectionPoint,
    Electricity_600VLL_3Ph_60HzInletConnectionPoint,
)
from ...connections.liquid import (
    ChilledWaterInletConnectionPoint,
    ChilledWaterOutletConnectionPoint,
    HotWaterInletConnectionPoint,
    HotWaterOutletConnectionPoint,
    WaterBidirectionalConnectionPoint,
    WaterInletConnectionPoint,
    WaterOutletConnectionPoint,
)
from ...connections.refrigerant import (
    RefrigerantBidirectionalConnectionPoint,
    RefrigerantInletConnectionPoint,
    RefrigerantOutletConnectionPoint,
)
from ...core import BOB, S223, Equipment, PropertyReference
from ...enum import (  # , R134a, R404a, R407c, R448a, R449a, R452a, R454b, R507a
    Refrigerant,
    Role,
)
from ...properties.force import Pressure
from ...properties.temperature import Temperature
from ...template import template_update

_namespace = BOB

coil_template = {
    "cp": {},
    "properties": {
        ("averageSurfaceTemperature", Temperature): {},
        ("internalPressure", Pressure): {},
        ("internalTemperature", Temperature): {},
    },
}


class Coil(Equipment):
    _class_iri = S223.Coil
    airInlet: AirInletConnectionPoint
    airOutlet: AirOutletConnectionPoint
    # Those could come from a valve, SCR, Triac, etc...
    modulation: PropertyReference
    onOffCommand: PropertyReference

    def __init__(self, config: Dict = coil_template, **kwargs):
        config["properties"] = config.get("properties", coil_template["properties"])
        kwargs = {**config.get("params", {}), **kwargs}
        super().__init__(config, **kwargs)
        self.airOutlet.paired_to(self.airInlet)


class WaterCoil(Coil):
    _class_iri = S223.Coil
    waterInlet: WaterInletConnectionPoint
    waterOutlet: WaterOutletConnectionPoint

    def __init__(self, config: Dict = coil_template, **kwargs):
        config["properties"] = config.get("properties", coil_template["properties"])
        kwargs = {**config.get("params", {}), **kwargs}
        super().__init__(config, **kwargs)
        self.waterOutlet.paired_to(self.waterInlet)


class DXCoolingCoil(Coil):
    _class_iri = S223.CoolingCoil
    refrigerantInlet: RefrigerantInletConnectionPoint
    refrigerantOutlet: RefrigerantOutletConnectionPoint

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update({}, config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        super().__init__(_config, **kwargs)
        self += Role.Cooling
        self.refrigerantOutlet.paired_to(self.refrigerantInlet)


class ChilledWaterCoil(Coil):
    _class_iri = S223.CoolingCoil
    chilledWaterInlet: ChilledWaterInletConnectionPoint
    chilledWaterOutlet: ChilledWaterOutletConnectionPoint

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update({}, config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        super().__init__(_config, **kwargs)
        self += Role.Cooling
        self.chilledWaterOutlet.paired_to(self.chilledWaterInlet)


class HotWaterCoil(Coil):
    _class_iri = S223.HeatingCoil
    hotWaterInlet: HotWaterInletConnectionPoint
    hotWaterOutlet: HotWaterOutletConnectionPoint

    def __init__(self, config: Dict = coil_template, **kwargs):
        _config = template_update({}, config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        super().__init__(_config, **kwargs)
        self += Role.Heating
        self.hotWaterOutlet.paired_to(self.hotWaterInlet)


class HeatpumpCoil(Coil):
    _class_iri = S223.Coil
    gasPortA: RefrigerantBidirectionalConnectionPoint
    gasPortB: RefrigerantBidirectionalConnectionPoint
    # airInlet: AirInletConnectionPoint
    # airOutlet: AirOutletConnectionPoint

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update(coil_template, config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        super().__init__(_config, **kwargs)
        self += Role.Heating
        self += Role.Cooling
        self.gasPortB.paired_to(self.gasPortA)

    def set_gas_type(self, gas: Refrigerant):
        self.set_medium(["gasPortA", "gasPortB"], gas)


# Electrical Coil
electricalheating_template = {
    "cp": {"electricalInlet": Electricity_600VLL_3Ph_60HzInletConnectionPoint},
    "properties": {
        ("amps", Amps): {},
        ("kW", ElectricPowerkW): {},
        ("modulation", PercentCommand): {},
        ("onOffCommand", OnOffCommand): {},
    },
}


class ElectricalHeatingCoil(Equipment):
    _class_iri = S223.ElectricResistanceElement
    airInlet: AirInletConnectionPoint
    airOutlet: AirOutletConnectionPoint
    # Those could come from a valve, SCR, Triac, etc...
    modulation: PropertyReference
    onOffCommand: PropertyReference

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update(electricalheating_template, config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        super().__init__(_config, **kwargs)
        self += Role.Heating
        self.airOutlet.paired_to(self.airInlet)


# Electrical Coil
electricalradiant_template = {
    "cp": {"electricalInlet": Electricity_240VLL_1Ph_60HzInletConnectionPoint},
    "properties": {
        ("amps", Amps): {},
        ("kW", ElectricPowerkW): {},
    },
}


# Baseboard, radiant panel, heating floor
class ElectricalRadiantHeatingCoil(Equipment):
    _class_iri = S223.RadiantHeater
    airContact: AirBidirectionalConnectionPoint

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update(electricalradiant_template, config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        super().__init__(_config, **kwargs)
        self += Role.Heating


# Water heaters
element_template = {
    "cp": {"electricalInlet": Electricity_240VLL_1Ph_60HzInletConnectionPoint},
    "properties": {
        ("amps", Amps): {},
        ("kW", ElectricPowerkW): {},
        ("modulation", PercentCommand): {},
        ("onOffCommand", OnOffCommand): {},
    },
}


class ImmersedResistanceHeaterElement(Equipment):
    _class_iri = S223.ElectricResistanceElement
    fluidContact: WaterBidirectionalConnectionPoint

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update(element_template, config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        super().__init__(_config, **kwargs)
        self += Role.Heating



from ..core import BOB, P223, QUANTITYKIND, UNIT, Medium, QuantifiableObservableProperty

_namespace = BOB


class AirChangePerHour(QuantifiableObservableProperty):
    """
    airflow / volume * 1hour


    ft^3           1
    ____   *   __________  *  60min = ACH
    min        vol (ft^3)

    """

    _class_iri = P223.AirChangePerHour
    hasQuantityKind = QUANTITYKIND.Dimensionless
    hasUnit = UNIT.UNITLESS


class Count(QuantifiableObservableProperty):
    _class_iri = BOB.Count
    hasQuantityKind = QUANTITYKIND.Dimensionless
    hasUnit = UNIT["NUM"]
    ofMedium: Medium



import logging
from typing import Dict

from rdflib import URIRef


from ...connections.air import AirInletConnectionPoint, AirOutletConnectionPoint
from ...connections.electricity import (
    ElectricalInletConnectionPoint,
)
from ...core import (
    BOB,
    S223,
    Equipment,
)
from ...template import configure_relations, template_update

# logging
_log = logging.getLogger(__name__)

# namespace
_namespace = BOB


fan_template = {"cp": {"electricalInlet": ElectricalInletConnectionPoint}}


class Fan(Equipment):
    """
    A fan is composed of a blower and an electrical motor
    """

    _class_iri: URIRef = S223.Fan
    # electricalInlet: ElectricalInletConnectionPoint # needs to be in a template so other templates can override it.
    airInlet: AirInletConnectionPoint
    airOutlet: AirOutletConnectionPoint

    def __init__(self, config: Dict = None, **kwargs):
        _config = template_update(fan_template, config=config)
        kwargs = {**_config.pop("params", {}), **kwargs}
        _log.info(f"Fan.__init__ {_config} {kwargs}")
        _relations = _config.pop("relations", [])
        super().__init__(_config, **kwargs)
        configure_relations(self, _relations)
        self.airOutlet.paired_to(self.airInlet)



from ..core import (
    P223,
    QUANTITYKIND,
    UNIT,
    Medium,
    Substance,
    QuantifiableObservableProperty,
)

_namespace = P223


class ParticulateCount(QuantifiableObservableProperty):
    hasQuantityKind = QUANTITYKIND.NumberDensity
    hasUnit = UNIT["NUM-PER-M3"]
    ofMedium: Medium
    ofSubstance: Substance



from ..core import BOB
from ..producer import Producer, ProducerInput, ProducerOutput

_namespace = BOB


class Causality(Producer):
    """
    Simple causality black box. Its goal is to express a
    simple relation between the output and the input.
    It tells : this input, causes an effect on that output.
    """

    cause_input: ProducerInput
    effect_output: ProducerOutput


class Differential(Producer):
    """
    This causality block goals explain that the output is the result
    of substracting high input and low input.
    """

    high_input: ProducerInput
    low_input: ProducerInput
    differential_output: ProducerOutput


class HandOffAutoMotorStarter(Producer):
    """
    Typically used in Electrical Motor Starter, where a selector allow manual, off, or auto
    The contactor position is commanded on or off depending on standard input (manual or auto + automatic command signal)
    Fire input, protection input and overload are protection that will turn off the contactor, stopping the load.
    """

    manual_or_off_input: ProducerInput  # front selector
    auto_input: ProducerInput  # from external source (ie. controller)
    fire_input: ProducerInput  # fire contact allowing the command to work
    protection_input: ProducerInput  # like freeze protection
    overload_input: ProducerInput  # Overloads from the contactor will make the starte go Off if current went too high
    contactor_output: ProducerOutput


class VFDProducer(Producer):
    """
    Very simple implementation of a VFD Producer.
    This is not Function, only relating different input properties with output properties.
    Any input causes an effect on all output in some way.

    NOTE : This is not the role of the producer to define in details the effects.
    """

    manual_or_off_input: ProducerInput  # front selector
    auto_input: ProducerInput  # from external source (ie. controller)
    speed_ref_input: ProducerInput
    fire_input: ProducerInput  # fire contact allowing the command to work
    run_enable_input: ProducerInput  # like freeze protection
    frequency_output: ProducerOutput
    voltage_output: ProducerOutput
    current_output: ProducerOutput
    torque_output: ProducerOutput
    rpm_output: ProducerOutput
    drive_running: ProducerOutput
